"""
These tools are used to identify areas that meet a number of different criteria you specify.

find_similar_locations finds locations most similar to one or more reference locations based on criteria you specify.
"""
import json as _json

import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from ._util import _id_generator, _feature_input, _set_context, _create_output_service, GAJob
import datetime
_log = _logging.getLogger(__name__)

# url = "https://dev003153.esri.com/gax/rest/services/System/GeoAnalyticsTools/GPServer"

_use_async = True

def geocode_locations(input_layer,
                      country=None,
                      category=None,
                      include_attributes=True,
                      locator_parameters=None,
                      output_name=None,
                      geocode_service=None,
                      geocode_parameters=None,
                      gis=None,
                      context=None,
                      future=False):
    """
    .. image:: _static/images/geocode_locations/geocode_locations.png 

    The ``geocode_locations`` task geocodes a table from a big data file share. 
    The task uses a geocode utility service configured with your portal. If you 
    do not have a geocode utility service configured, talk to your administrator. 
    `Learn more about configuring a locator 
    service <https://enterprise.arcgis.com/en/portal/latest/administer/windows/configure-portal-to-geocode-addresses.htm>`_.

    When preparing to use the Geocode Location task be sure to review `Best Practices 
    for geocoding with GeoAnalytics Server <https://enterprise.arcgis.com/en/portal/latest/use/geoanalytics-geocoding-best-practices.htm>`_.

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  Required layer. The tabular input that will be geocoded. See :ref:`Feature Input<FeatureInput>`.
    --------------------------   ---------------------------------------------------------------
    country                      Optional string. If all your data is in one country, this helps
                                 improve performance for locators that accept that variable.
    --------------------------   ---------------------------------------------------------------
    category                     Optional string. Enter a category for more precise geocoding
                                 results, if applicable. Some geocoding services do not support
                                 category, and the available options depend on your geocode service.
    --------------------------   ---------------------------------------------------------------
    include_attributes           Optional boolean. A Boolean value to return the output fields 
                                 from the geocoding service in the results. To output all available 
                                 output fields, set this value to 'True'. Setting the value to false will 
                                 return your original data and with geocode coordinates. Some geocoding 
                                 services do not support output fields, and the available options depend 
                                 on your geocode service.
    --------------------------   ---------------------------------------------------------------
    locator_parameters           Optional dict. Additional parameters specific to your locator.
    --------------------------   ---------------------------------------------------------------
    output_name                  Optional string. The task will create a feature service of the
                                 results. You define the name of the service.
    --------------------------   ---------------------------------------------------------------
    geocode_service              Optional string or Geocoder. The URL of the geocode service 
                                 that you want to geocode your addresses against. The URL must end in 
                                 geocodeServer and allow batch requests. The geocode service must be 
                                 configured to allow for batch geocoding. For more information, 
                                 see `Configuring batch geocoding <https://enterprise.arcgis.com/en/portal/latest/administer/windows/configure-portal-to-geocode-addresses.htm>`_
    --------------------------   ---------------------------------------------------------------
    geocode_parameters           optional dict. This includes parameters that help parse
                                 the input data, as well the field lengths and a field mapping.
                                 This value is the output from the AnalyzeGeocodeInput tool
                                 available on your server designated to geocode. It is important
                                 to inspect the field mapping closely and adjust them accordingly
                                 before submitting your job, otherwise your geocoding results may
                                 not be accurate. It is recommended to use the output from
                                 AnalyzeGeocodeInput and modify the field mapping instead of
                                 constructing this JSON by hand.

                                 **Values**

                                 **field_info** - A list of triples with the field names of your input
                                 data, the field type (usually TEXT), and the allowed length
                                 (usually 255).
                                 Example: [['ObjectID', 'TEXT', 255], ['Address', 'TEXT', 255],
                                          ['Region', 'TEXT', 255], ['Postal', 'TEXT', 255]]

                                 **header_row_exists** - Enter true or false.

                                 **column_names** - Submit the column names of your data if your data
                                 does not have a header row.

                                 **field_mapping** - Field mapping between each input field and
                                 candidate fields on the geocoding service.
                                 Example: [['ObjectID', 'OBJECTID'], ['Address', 'Address'],
                                          ['Region', 'Region'], ['Postal', 'Postal']]
    --------------------------   ---------------------------------------------------------------
    gis                          Optional GIS. The GIS on which this tool runs. If not
                                 specified, the active GIS is used.
    --------------------------   ---------------------------------------------------------------
    context                      Optional dict. Context contains additional settings that affect task execution. 
                                 For this task, there are three settings:

                                 Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                 Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                 Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    --------------------------   ---------------------------------------------------------------
    future                       Optional boolean. If True, a GPJob is returned instead of
                                 results. The GPJob can be queried on the status of the execution.
    ==========================   ===============================================================


    :returns: Feature Layer

    .. code-block:: python

            # Usage Example: To geocode a big data file share of mailing addresses in the United States Northwest.

            geocode_server = "https://mymachine.domain.com/server/rest/services/USALocator/GeocodeServer"
            geo_parameters = {"field_info": "[('ObjectID', 'TEXT', 255), ('Street', 'TEXT', 255), ('City', 'TEXT', 255), ('Region', 'TEXT', 255), ('State', 'TEXT', 255)]", "column_names": "", "file_type": "table", "header_row_exists": "true", "field_mapping": "[[\"Street\", \"Street\"], [\"City\", \"City\"], [\"State\", \"State\"], [\"ZIP\", \"ZIP\"]]"}
            geocode_result = find_locations.geocode_locations(input_layer=NW_addresses, 
                                                        output_name="geocoded_NW_USA",
                                                        geocode_service=geocode_server,
                                                        geocode_parameters = geo_parameters)

    """
    from arcgis.features.layer import Layer
    from arcgis.gis import Item
    from arcgis.geocoding._functions import Geocoder
    kwargs = locals()
    tool_name = "GeocodeLocations"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value
    if output_name is None:
        output_service_name = 'Geocoding_Results_' + _id_generator()
        output_service_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    if isinstance(input_layer, str):
        input_layer = {'url' : input_layer}
    elif isinstance(input_layer, Item):
        input_layer = input_layer.layers[0]._lyr_dict
        if 'type' in input_layer:
            input_layer.pop('type')
    elif isinstance(input_layer, Layer):
        input_layer = input_layer._lyr_dict
        if 'type' in input_layer:
            input_layer.pop('type')
    elif isinstance(input_layer, dict) and \
         not "url" in input_layer:
        raise ValueError("Invalid Input: input_layer dictionary" + \
                         " must have format {'url' : <url>}")
    elif isinstance(input_layer, dict) and "url" in input_layer:
        pass
    else:
        raise ValueError("Invalid input_layer input. Please pass an Item, " + \
                         "Big DataStore Layer or Big DataStore URL to geocode.")

    if geocode_service is None:
        for service in gis.properties.helperServices.geocode:
            if 'batch' in service and service['batch'] == True:
                geocode_service_url = service["url"]
                break
        if geocode_service_url is None:
            raise ValueError("A geocoder with batch enabled must be configured" + \
                             " with this portal to use this service.")
        params['geocode_service_url'] = geocode_service_url
    elif isinstance(geocode_service, Geocoder):
        geocode_service = geocode_service.url
        params['geocode_service_url'] = geocode_service_url
    elif isinstance(geocode_service, str):
        params['geocode_service_url'] = geocode_service
    else:
        raise ValueError("geocode_service_url must be a string or GeoCoder")

    if geocode_parameters is None:
        from arcgis.geoprocessing._tool import Toolbox
        analyze_geocode_url = gis.properties.helperServices.asyncGeocode.url
        tbx = Toolbox(url=analyze_geocode_url, gis=gis)
        geocode_parameters = tbx.analyze_geocode_input(input_table=input_layer,
                                                       geocode_service_url=geocode_service_url)
        params['geocode_parameters'] = geocode_parameters
    output_service = _create_output_service(gis, output_name,
                                            output_service_name, 'Geocoded Locations')
    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "geocode_service_url": (str, "geocodeServiceURL"),
        "geocode_parameters": (str, "geocodeParameters"),
        "country": (str, "sourceCountry"),
        "category": (str, "category"),
        "include_attributes" : (bool, "includeAttributes"),
        "locator_parameters" : (str, "locatorParameters"),
        "output_name": (str, "outputName"),
        "output": (_FeatureSet, "output"),
        "context": (str, "context")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        res = _execute_gp_tool(gis, tool_name, params, param_db,
                               return_values, _use_async, url, True,
                               future=future)
        return output_service
    except:
        output_service.delete()
        raise
    return

def detect_incidents(input_layer,
                     track_fields,
                     start_condition_expression,
                     end_condition_expression=None,
                     output_mode="AllFeatures",
                     time_boundary_split=None,
                     time_split_unit=None,
                     time_reference=None,
                     output_name=None,
                     gis=None,
                     context=None,
                     future=False):
    """

    .. image:: _static/images/detect_incidents/detect_incidents.png 

    The ``detect_incidents`` task works with a time-enabled layer of points,
    lines, areas, or tables that represents an instant in time. Using
    sequentially ordered features, called tracks, this tool determines
    which features are incidents of interest. Incidents are determined by
    conditions that you specify. First, the tool determines which features
    belong to a track using one or more fields. Using the time at each
    feature, the tracks are ordered sequentially and the incident condition
    is applied. Features that meet the starting incident condition are
    marked as an incident. You can optionally apply an ending incident
    condition; when the end condition is 'True', the feature is no longer
    an incident. The results will be returned with the original features
    with new columns representing the incident name and indicate which
    feature meets the incident condition. You can return all original
    features, only the features that are incidents, or all of the features
    within tracks where at least one incident occurred.

    For example, suppose you have GPS measurements of hurricanes every 10
    minutes. Each GPS measurement records the hurricane's name, location,
    time of recording, and wind speed. Using these fields, you could create
    an incident where any measurement with a wind speed greater than 208
    km/h is an incident titled Catastrophic. By not setting an end
    condition, the incident would end if the feature no longer meets the
    start condition (wind speed slows down to less than 208).

    Using another example, suppose you were monitoring concentrations of a
    chemical in your local water supply using a field called
    contanimateLevel. You know that the recommended levels are less than
    0.01 mg/L, and dangerous levels are above 0.03 mg/L. To detect
    incidents, where a value above 0.03mg/L is an incident, and remains an
    incident until contamination levels are back to normal, you create an
    incident using a start condition of contanimateLevel > 0.03 and an end
    condition of contanimateLevel < 0.01. This will mark any sequence where
    values exceed 0.03mg/L until they return to a value less than 0.01.

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  Required layer. The table, point, line or polygon features 
                                 containing potential incidents. See :ref:`Feature Input<FeatureInput>`.
    --------------------------   ---------------------------------------------------------------
    track_fields                 Required string. The fields used to identify distinct tracks.
                                 There can be multiple ``track_fields``.
    --------------------------   ---------------------------------------------------------------
    start_condition_expression   Required string. The condition used to identify incidents. If there
                                 is no ``end_condition_expression`` specified, any feature
                                 that meets this condition is an incident. If there
                                 is an end condition, any feature that meets the
                                 ``start_condition_expression`` and does not meet the
                                 ``end_condition_expression`` is an incident.
                                 The expressions are Arcade expressions.
    --------------------------   ---------------------------------------------------------------
    end_condition_expression     Optional string. The condition used to identify incidents. If there is
                                 no ``end_condition_expression`` specified, any feature that
                                 meets this condition is an incident. If there is an
                                 end condition, any feature that meets the
                                 ``start_condition_expression`` and does not meet the
                                 ``end_condition_expression`` is an incident. This is an
                                 Arcade expression.
    --------------------------   ---------------------------------------------------------------
    output_mode                  Optional string. Determines which features are returned. 

                                 Choice list: [AllFeatures', 'Incidents']

                                 - ``AllFeatures`` - All of the input features are returned.
                                 - ``Incidents`` - Only features that were found to be incidents
                                   are returned.

                                 The default value is 'AllFeatures'.
    --------------------------   ---------------------------------------------------------------
    time_boundary_split          Optional integer. A time boundary to detect and incident. A time 
                                 boundary allows your to analyze values within a defined time span. 
                                 For example, if you use a time boundary of 1 day, starting on January 
                                 1st, 1980 tracks will be analyzed 1 day at a time. The time boundary 
                                 parameter was introduced in ArcGIS Enterprise 10.7.

                                 The ``time_boundary_split`` parameter defines the scale of the time boundary. 
                                 In the case above, this would be 1. See the portal documentation for 
                                 this tool to learn more.
    --------------------------   ---------------------------------------------------------------
    time_split_unit              Optional string. The unit to detect an incident is `time_boundary_split` is used.

                                 Choice list: ['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds'].
    --------------------------   ---------------------------------------------------------------
    time_reference               Optional datetime.detetime. The starting date/time where analysis will
                                 begin from.
    --------------------------   ---------------------------------------------------------------
    output_name                  optional string, The task will create a feature service of the
                                 results. You define the name of the service.
    --------------------------   ---------------------------------------------------------------
    gis                          optional GIS, the GIS on which this tool runs. If not
                                 specified, the active GIS is used.
    --------------------------   ---------------------------------------------------------------
    context                      Optionl dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                 #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                 #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                 #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                 #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    --------------------------   ---------------------------------------------------------------
    future                       optional boolean. If True, a GPJob is returned instead of
                                 results. The GPJob can be queried on the status of the execution.

                                 The default value is 'False'.
    ==========================   ===============================================================

    :returns: result_layer : Output Features as feature layer collection item.

    .. code-block:: python

            # Usage Example: This example finds when and where snowplows were moving slower than 10 miles per hour by calculating the mean of a moving window of five speed values.
            
            arcgis.env.verbose = True # set environment
            arcgis.env.defaultAggregations = True # set environment

            delay_incidents = output = detect_incidents(input_layer=snowplows, 
                                                        track_fields="plowID, dayOfYear", 
                                                        start_condition_expression="Mean($track.field["speed"].window(-5, 0)) < 10", 
                                                        output_name="Slow_Plow_Incidents")
    """
    kwargs = locals()
    tool_name = "DetectIncidents"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Detect_Incidents_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Detect Track Incidents')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "track_fields": (str, "trackFields"),
        "start_condition_expression": (str, "startConditionExpression"),
        "end_condition_expression": (str, "endConditionExpression"),
        "output_mode": (str, "outputMode"),
        "time_boundary_split" : (int, "timeBoundarySplit"),
        "time_split_unit" : (str, "timeBoundarySplitUnit"),
        "time_reference" : (datetime.datetime, "timeBoundaryReference"),
        "output_name": (str, "outputName"),
        "output": (_FeatureSet, "output"),
        "context": (str, "context")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

    return


def find_similar_locations(
                           input_layer,
                           search_layer,
                           analysis_fields,
                           most_or_least_similar="MostSimilar",
                           match_method="AttributeValues",
                           number_of_results=10,
                           append_fields=None,
                           output_name=None,
                           gis=None,
                           context=None,
                           future=False,
                           return_tuple=False):
    """
    .. image:: _static/images/find_similar_locations/find_similar_locations.png 

    The ``find_similar_locations`` task measures the similarity of candidate locations to one or more reference locations.

    Based on criteria you specify, ``find_similar_locations`` can answer questions such as the following:

        * Which of your stores are most similar to your top performers with regard to customer profiles?
        * Based on characteristics of villages hardest hit by the disease, which other villages are high risk?
        * To answer questions such as these, you provide the reference locations (the ``input_layer`` parameter), 
          the candidate locations (the ``search_layer`` parameter), and the fields representing the criteria 
          you want to match. For example, the ``input_layer`` might be a layer containing your top performing stores 
          or the villages hardest hit by the disease. The ``search_layer`` contains your candidate locations to search. 
          This might be all of your stores or all other villages. Finally, you supply a list of fields to use for 
          measuring similarity. The ``find_similar_locations`` task will rank all of the candidate locations by how 
          closely they match your reference locations across all of the fields you have selected.
 
    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  Required layer. The ``input_layer`` contains one or more reference locations 
                                 against which features in the ``search_layer`` will be evaluated for similarity. 
                                 For example, the ``input_layer`` might contain your top performing stores or the 
                                 villages hardest hit by a disease. See :ref:`Feature Input<FeatureInput>`.
                                  
                                 It is not uncommon for ``input_layer`` and ``search_layer`` to be the same feature service. 
                                 For example, the feature service contains locations of all stores, one of which 
                                 is your top performing store. If you want to rank the remaining stores from most 
                                 to least similar to your top performing store, you can provide a filter for both 
                                 ``input_layer`` and ``search_layer``. The filter on ``input_layer`` would select the top performing 
                                 store, while the filter on ``search_layer`` would select all stores except for the top 
                                 performing store. You can use the optional filter parameter to specify reference locations.

                                 If there is more than one reference location, similarity will be based on averages 
                                 for the fields you specify in the ``analysis_fields`` parameter. For example, if there 
                                 are two reference locations and you are interested in matching population, the task 
                                 will look for candidate locations in ``search_layer`` with populations that are most 
                                 like the average population for both reference locations. If the values for the 
                                 reference locations are 100 and 102, for example, the task will look for candidate 
                                 locations with populations near 101. Consequently, you will want to use fields for 
                                 the reference locations fields that have similar values. If, for example, the 
                                 population values for one reference location is 100 and the other is 100,000, 
                                 the tool will look for candidate locations with population values near the average 
                                 of those two values: 50,050. Notice that this averaged value is nothing like the 
                                 population for either of the reference locations. 
    --------------------------   ---------------------------------------------------------------  
    search_layer                 Required layer. The layer containing candidate locations that will be 
                                 evaluated against the reference locations.  See :ref:`Feature Input<FeatureInput>`.
    --------------------------   ---------------------------------------------------------------      
    analysis_fields              Required string. A list of fields whose values are used to determine similarity. 
                                 They must be numeric fields, and the fields must exist on both the ``input_layer`` 
                                 and the ``search_layer``. Depending on the ``match_method`` selected, the task will 
                                 find features that are most similar based on values or profiles of the fields.
    --------------------------   ---------------------------------------------------------------      
    most_or_least_similar        Optional string. The features you want to be returned. You can search for 
                                 features that are either most similar or least similar to the ``input_layer``, 
                                 or search both the most and least similar.
                                 
                                 Choice list:['MostSimilar', 'LeastSimilar', 'Both']

                                 The default value is 'MostSimilar'.
    --------------------------   ---------------------------------------------------------------
    match_method                 Optional string. The method you select determines how matching is determined. 

                                 Choice list:['AttributeValues', 'AttributeProfiles']    

                                    * The ``AttributeValues`` method uses the squared differences of standardized values. 
                                    * The ``AttributeProfiles`` method uses cosine similarity mathematics to compare the profile 
                                      of standardized values. Using ``AttributeProfiles`` requires the use of at least two analysis fields.   

                                 The default value is 'AttributeValues'.
    --------------------------   ---------------------------------------------------------------    
    number_of_results            Optional integer. The number of ranked candidate locations output 
                                 to ``similar_result_layer``. If ``number_of_results`` is not set, the 10 
                                 locations will be returned. The maximum number of results is 10000.

                                 The default value is 10.
    --------------------------   ---------------------------------------------------------------   
    append_fields                Optional string. Optionally add fields to your data from your search layer. 
                                 By default, all fields from the search layer are appended.
    --------------------------   ---------------------------------------------------------------      
    output_name                  Optional string. The task will create a feature service of the results. 
                                 You define the name of the service.
    --------------------------   ---------------------------------------------------------------      
    gis                          Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    --------------------------   ---------------------------------------------------------------      
    context                      Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                 #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                 #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                 #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                 #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    --------------------------   ---------------------------------------------------------------      
    future                       Optional boolean. If 'True', a GPJob is returned instead of results. The GPJob can be queried on the status of the execution.

                                 The default value is 'False'.
    --------------------------   ---------------------------------------------------------------      
    return_tuple                 Optional boolean. If 'True', a named tuple with multiple output keys is returned.
                                 
                                 The default value is 'False'. 
    ==========================   ===============================================================

    :returns: named tuple with the following keys if ``return_tuple`` is set to 'True':

      "output" : feature layer

      "process_info" : list

    else returns a feature layer of the results.

    .. code-block:: python

            # Usage Example: To find potential retail locations based on the current top locations and their attributes.
            
            similar_location_result = find_similar_locations(input_layer=stores_layer,
                                                             search_layer=locations,
                                                             analysis_fields="median_income, population, nearest_competitor",
                                                             most_or_least_similar="MostSimilar", 
                                                             match_method="AttributeValues", 
                                                             number_of_results=50, 
                                                             output_name="similar_locations")
    """  
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Similar Locations_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Find Similar Locations')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "search_layer": (_FeatureSet, "searchLayer"),
        "analysis_fields": (str, "analysisFields"),
        "most_or_least_similar": (str, "mostOrLeastSimilar"),
        "match_method": (str, "matchMethod"),
        "number_of_results": (int, "numberOfResults"),
        "append_fields": (str, "appendFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "return_tuple": (bool, "returnTuple"),        
        "output": (_FeatureSet, "Output Features"),
        "process_info": (list, "processInfo")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
        {"name": "process_info", "display_name": "Process Information", "type": list}
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, "FindSimilarLocations", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        res = _execute_gp_tool(gis, "FindSimilarLocations", params, param_db, return_values, _use_async, url, True, future=future)
        
        if return_tuple:
            return res
        else:
            return output_service   
    except:
        output_service.delete()
        raise 

find_similar_locations.__annotations__ = {
    'most_or_least_similar': str,
    'match_method': str,
    'number_of_results': int,
    'append_fields': str,
    'output_name': str}




import time as _time
from arcgis.gis import GIS
from arcgis._impl import _ArcGISConnection
from arcgis._impl.common._mixins import PropertyMap
from arcgis.gis import Item
###########################################################################
class PortalDataStore(object):
    """

    The datastores resource page provides access to operations that allow you to do the
    following:

    - Validate a data store against your server.
    - Register a data store to your server.
    - Retrieve a list of servers your data store is registered to.
    - Refresh your data store registration information in the server.
    - Unregister a data store from your server.

    Additionally, users also have the ability to bulk publish layers from a data store,
    retrieve a list of previously published layers, and delete bulk-published layers.

    """
    _con = None
    _gis = None
    _url = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        self._gis = gis
        self._con = gis._con
    #----------------------------------------------------------------------
    def __str__(self):
        return "< PortalDataStore @ {url} >".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "< PortalDataStore @ {url} >".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the datastore"""
        if self._properties is None:
            params = {'f' : 'json'}
            res = self._con.get(self._url, params)
            self._properties = PropertyMap(res)
        return self._properties
    #----------------------------------------------------------------------
    def register(self, item, server_id):
        """

        The `register` method allows for Data Store type Items to be added to an ArcGIS Server instance.
        Before registering a data store, it is recommended that you validate your data store with the
        given server.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item/Item Id (as string). The item Id or Item of the data
                               store to register with the server. Note that a data store can be
                               registered on multiple servers.
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique id of the server you want to register
                               the datastore with.
        ==================     ====================================================================

        :returns: Boolean

        """
        if isinstance(item, Item):
            item_id = item.id
        elif isinstance(item, str):
            item_id = item
        url = "{base}/addToServer".format(base=self._url)
        params = {
            'f' : 'json',
            'datastoreId' : item_id,
            'serverId' : server_id
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    @property
    def _all_datasets(self):
        """
        The _all_datasets resource page provides access to bulk publishing
        operations. These operations allow users to publish and synchronize
        datasets from a given data store, return a list of published layers,
        and remove all previously published layers in preparation for the
        deregistration of a data store.

        :returns: Boolean

        """
        url = "{base}/allDatasets".format(base=self._url)
        params = {
            'f' : 'json'}
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def delete_layers(self, item):
        """
        Before a data store can be unregistered from a server, all of its
        bulk-published layers must be deleted. The delete_layers removes all
        layers published from the data store.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The Data Store `Item` to delete all published layers.
        ==================     ====================================================================

        :returns: Boolean

        """
        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)
        params = {
            'f' : 'json',
            'datastoreId' : item_id
        }
        url = "{base}/allDatasets/deleteLayers"
        res = self._con.post(url, params)
        if res['success'] == True:
            status = item.status(self, job_id=res['jobId'])
            while status["status"].lower() != 'completed':
                status = item.status(self, job_id=res['jobId'])
                if status['status'].lower() == "failed":
                    return False
                else:
                    _time.sleep(2)
            return True
        return False
    #----------------------------------------------------------------------
    def layers(self, item):
        """
        The `layers operation returns a list of layers bulk published from a
        data store with the `publish_layers` method. The `layers` method
        returns an array of tuples, with each tuple containing two
        objects: a layer and the dataset it was published from.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The Data Store `Item` to list all published layers
                               and registered datasets.
        ==================     ====================================================================

        :returns: List

        """

        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)

        url = "{base}/allDatasets/getLayers".format(base=self._url)
        params = {
            'f' : 'json',
            'datastoreId' : item_id
        }
        res = self._con.post(url, params)
        if "layerAndDatasets" in res:
            return res["layerAndDatasets"]
        return []
    #----------------------------------------------------------------------
    def servers(self, item):
        """
        The `servers` property returns a list of your servers that a given
        data store has been registered to. This operation returns the
        serverId, the server name, both the server and admin URLs, and
        whether or not the server is hosted.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The Data Store `Item` to list all registered servers.
        ==================     ====================================================================

        :returns: List

        """

        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)

        url = self._url + "/getServers"
        params = {
            'f' : 'json',
            'datastoreId' : item_id
        }
        res = self._con.post(url, params)
        if 'servers' in res:
            return res['servers']
        return res
    #----------------------------------------------------------------------
    def publish_layers(self, item, srv_config, server_id, folder=None, server_folder=None):
        """
        The `publish_layers` operation publishes, or syncs, the datasets from a
        data store onto your ArcGIS Server, resulting in at least one layer per
        dataset.

        When this operation is called for the first time, every parameter in
        the operation must be passed through. On subsequent calls,
        publish_layers will synchronize the datasets and created layers, which
        includes both publishing layers from new datasets and removing layers
        for datasets no longer found in the data store.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item. The Data Store `Item` to publish from.
        ------------------     --------------------------------------------------------------------
        srv_config             Required Dict. The JSON that will be used as a template for all the
                               services that will be published or synced. This JSON can be used to
                               change the properties of the data store's map and feature services.
                               Only map service configurations with feature services enabled are
                               supported by this parameter.
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The serverId that the datasets will be published to.
        ------------------     --------------------------------------------------------------------
        folder                 Optional String. The folder to which the datasets will be published.
        ------------------     --------------------------------------------------------------------
        server_folder          Optional String. The name of the server folder.
        ==================     ====================================================================

        :returns: Boolean

        """
        if isinstance(item, Item):
            item_id = item.id
        else:
            item_id = item
            item = self._gis.content.get(item_id)
        url = self._url + "/allDatasets/publishLayers"
        params = {
            'f' : 'json',
            'datastoreId' : item_id,
            'templateScvConfig' : srv_config,
            'portalFolderId' : folder,
            'serverId' : server_id,
            'serverFolder' : server_folder,
        }
        res = self._con.post(url, params)
        if res['success'] == True:
            status = item.status(self, job_id=res['jobId'])
            while status["status"].lower() != 'completed':
                status = item.status(self, job_id=res['jobId'])
                if status['status'].lower() == "failed":
                    return False
                else:
                    _time.sleep(2)
            return True
        return False
    #----------------------------------------------------------------------
    def unregister(self, item, server_id):
        """
        Removes the datastore association from a server.

        :returns: Boolean

        """
        if isinstance(item, Item):
            item_id = item.id
        elif isinstance(item, str):
            item_id = item
        url = self._url + "/removeFromServer"
        params = {
            'f' : 'json',
            'datastoreId' : item_id,
            'serverId' : server_id
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def refresh_server(self, item, server_id):
        """
        After a data store has been registered, there may be times in which
        the data store's registration information may be changed. When
        changes like these occur, the server will need to be updated with
        the newly configured information so that your users will still be
        able to access the data store items without interruption. The
        `refresh_server` can be called to propagate these changes to your
        ArcGIS Server. This operation can only be performed after the data
        store information has been updated.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required Item/Item Id (as string). The item Id or Item of the data
                               store to register with the server. Note that a data store can be
                               registered on multiple servers.
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique id of the server you want to register
                               the datastore with.
        ==================     ====================================================================

        :returns: Boolean

        """
        if isinstance(item, Item):
            item_id = item.id
        elif isinstance(item, str):
            item_id = item

        url = self._url + "/refreshServer"
        params = {
            'f' : 'json',
            'datastoreId' : item_id,
            'serverId' : server_id
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def validate(self, server_id, item=None, config=None):
        """
        The `validate` ensures that your ArcGIS Server can connect and use
        the datasets stored within a given data store. While this operation
        can be called before or after the data store has been registered
        with your server, it is recommended that the validate operation is
        performed beforehand. A data store can be validated by using either
        its datastoreId or the JSON for an unregistered data store.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        server_id              Required String. The unique id of the server you want to register
                               the datastore with.
        ------------------     --------------------------------------------------------------------
        item                   Optional Item/Item Id (as string). The item Id or Item of the data
                               store to register with the server. Note that a data store can be
                               registered on multiple servers.
        ------------------     --------------------------------------------------------------------
        config                 Optional dict. The connection information for a new datastore.
        ==================     ====================================================================

        :returns: Boolean

        """

        if item and isinstance(item, Item):
            item_id = item.id
        elif item and isinstance(item, str):
            item_id = item
        else:
            item_id = None

        url = self._url + "/validate"
        params = {
                'f' : 'json',
                'serverId' : server_id
            }
        if item:
            params['datastoreId'] = item_id
        elif config:
            import json
            params['datastore'] = json.dumps(config)
        else:
            raise ValueError("Invalid parameters, an item or config is required.")
        res = self._con.post(url, params)
        if 'status' in res:
            if res['status'] == 'success':
                return True
            return res['status']
        return res



import os, csv
from warnings import warn

HAS_OPENCV = True
HAS_FASTAI = True
HAS_ARCPY = True

try:
    from .models import _tracker_util
    from fastprogress import progress_bar
    from ._image_utils import _draw_predictions
    from fastai.vision.image import bb2hw, Image, pil2tensor
except Exception:
    HAS_FASTAI = False

try:
    import cv2
except Exception:
    HAS_OPENCV = False

try:
    import arcpy
except Exception:
    HAS_ARCPY = False

class VideoUtils():

    @staticmethod    
    def predict_video(
        model,
        input_video_path,
        metadata_file, 
        threshold=0.5, 
        nms_overlap=0.1, 
        track=False, 
        visualize=False, 
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={'assignment_iou_thrd': 0.3, 'vanish_frames': 40, 'detect_frames': 10},
        visual_options={'show_scores': True, 'thickness': 2, 'fontface': 0, 'show_labels': True, 'color': (255, 255, 255)}
    ):
        if not HAS_OPENCV:
            raise Exception("This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24")

        video_read = cv2.VideoCapture(input_video_path)
        fps = video_read.get(cv2.CAP_PROP_FPS)
        video_obj = None
        success = True
        total_frames = int(video_read.get(cv2.CAP_PROP_FRAME_COUNT))
        frame_number = 0
        vmtis = ['vmtilocaldataset']

        object_id_mapping = {}
        object_id = 1
        tracker_list = []
        tracker_ind = 0

        thickness = None
        fontface = None
        show_labels = None
        color = None
        show_scores = True

        if visualize:
            thickness = visual_options.get('thickness', 2)
            fontface = visual_options.get('fontface', 0)
            show_labels = visual_options.get('show_labels', True)
            color = visual_options.get('color', (255, 255, 255))
            show_scores = visual_options.get('show_scores', True)

        for pb in progress_bar(range(total_frames)):
            success, frame = video_read.read()
            frame_number = frame_number + 1

            if not success and frame_number < total_frames:
                continue
            elif not success:
                break

            height, width, _ = frame.shape
            if visualize and not video_obj:
                if not output_file_path:
                    output_file_path = os.path.join(
                        os.path.dirname(input_video_path),
                        os.path.basename(input_video_path).split('.')[0] + '_predictions.avi'
                    )
                video_obj = cv2.VideoWriter(output_file_path, cv2.VideoWriter_fourcc(*'DIVX'), fps, (width, height))

            predictions, labels, scores = model.predict(frame, threshold=threshold, nms_overlap=nms_overlap,
                                                       return_scores=True)
            vmti_detections = '\n'

            if predictions:
                if track:
                    bboxes = []
                    for prediction in predictions:
                        bboxes.append(
                            [prediction[1], prediction[0], prediction[1] + prediction[3], prediction[0] + prediction[2]])

                    predictions, labels, scores, tracker_list, tracker_ind = _tracker_util.main_tracker(frame,
                                                                 bboxes,
                                                                 scores,
                                                                 tracker_options['assignment_iou_thrd'],
                                                                 tracker_options['vanish_frames'],
                                                                 tracker_options['detect_frames'], 
                                                                 tracker_list, 
                                                                 tracker_ind)

                    for index, data in enumerate(predictions):
                        top_left = max(0, (int(data[1]) - 1)) * width + int(data[0])
                        bottom_right = max(0, (int(data[1] + data[3]) - 1)) * width + int(data[0] + data[2])
                        center_pixel = (int(data[1]) + int((data[3]) / 2)) * width + (int(data[0]) + int((data[2]) / 2))

                        vmti_detections = f'{labels[index]} {scores[index] * 100} {top_left} {bottom_right} {center_pixel};' + vmti_detections
                else:
                    for index, data in enumerate(predictions):
                        top_left = max(0, (int(data[1]) - 1)) * width + int(data[0])
                        bottom_right = max(0, (int(data[1] + data[3]) - 1)) * width + int(data[0] + data[2])
                        center_pixel = (int(data[1]) + int((data[3]) / 2)) * width + (
                                int(data[0]) + int((data[2]) / 2))

                        if not object_id_mapping.get(labels[index]):
                            object_id_mapping[labels[index]] = object_id
                            object_id = object_id + 1

                        vmti_detections = f'{object_id_mapping[labels[index]]} {scores[index] * 100} {top_left} {bottom_right} {center_pixel};' + vmti_detections

                image = _draw_predictions(
                    frame,
                    predictions,
                    labels,
                    scores=scores,
                    show_scores=show_scores,
                    thickness=thickness,
                    fontface=fontface,
                    color=color,
                    show_labels=show_labels
                )
            else:
                image = frame

            vmtis.append(vmti_detections)

            if visualize:
                video_obj.write(image)

        if video_obj:
            video_obj.release()

        video_read.release()

        data = []
        index = 0

        file_exists = True
        fields = []

        if not os.path.exists(metadata_file):
            file_exists = False
            for vmti in vmtis:
                data.append([vmti])
        else:
            with open(metadata_file, 'r') as csvinput:
                for row in csv.reader(csvinput):
                    if index == 0:
                        fields = row
                    if len(vmtis) <= index:
                        data.append(row + [""])
                    else:
                        data.append(row + [vmtis[index]])
                    index = index + 1

        if 'vmtilocaldataset' in fields:
            warn("Field 'vmtilocaldataset' already exists in the file, appending column at the end.")

        if len(data) < len(vmtis):
            warn(f"Writing {len(data)} rows only!")

        with open(metadata_file, 'w', newline='') as csvoutput:
            writer = csv.writer(csvoutput)
            for row in data:
                writer.writerow(row)

        if not multiplex:
            return

        if not HAS_ARCPY:
            warn("Arcpy doesn't exist, multiplexing skipped.")
            return

        if not file_exists:
            warn("Metadata file doesn't exist, multiplexing skipped.")
            return

        if not multiplex_file_path:
            multiplex_file_path = os.path.join(
                os.path.dirname(input_video_path),
                os.path.basename(input_video_path).split('.')[0] + '_multiplex.MOV'
            )

        arcpy.ia.VideoMultiplexer(input_video_path, metadata_file, multiplex_file_path)



import json, tempfile
from pathlib import Path
from ._codetemplate import image_classifier_prf
from ._arcgis_model import _EmptyData
from functools import partial
import math

try:
    from ._arcgis_model import ArcGISModel, SaveModelCallback, _set_multigpu_callback
    import torch
    from torchvision import models
    from fastai.vision.learner import unet_learner, cnn_config
    import numpy as np
    from ._unet_utils import is_no_color, LabelCallback, _class_array_to_rbg
    from fastai.callbacks import EarlyStoppingCallback
    from torch.nn import Module as NnModule
    HAS_FASTAI = True
except Exception as e:
    class NnModule():
        pass
    HAS_FASTAI = False


def accuracy(input, target, void_code=0, class_mapping=None):  
    target = target.squeeze(1)
    mask = target != void_code
    return (input.argmax(dim=1)[mask] == target[mask]).float().mean()

class UnetClassifier(ArcGISModel):
    """
    Creates a Unet like classifier based on given pretrained encoder.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `UnetClassifier`, which
                            is `resnet34` by default.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :returns: `UnetClassifier` Object
    """

    def __init__(self, data, backbone=None, pretrained_path=None):

        super().__init__(data, backbone)

        self._code = image_classifier_prf

        backbone_cut = None
        backbone_split = None

        _backbone = self._backbone
        if hasattr(self, '_backbone_'):
            _backbone = self._backbone_
            
        if not (self._check_backbone_support(_backbone)):
            raise Exception(f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

        if hasattr(self, '_backbone_'):
            _backbone_meta = cnn_config(self._backbone_)
            backbone_cut = _backbone_meta['cut']
            backbone_split = _backbone_meta['split']

        acc_metric = partial(accuracy, void_code=0, class_mapping=data.class_mapping) 
        self.learn = unet_learner(data, arch=self._backbone, metrics=acc_metric, wd=1e-2, bottle=True, last_cross=True, cut=backbone_cut, split_on=backbone_split)
        self._arcgis_init_callback() # make first conv weights learnable
        self.learn.callbacks.append(LabelCallback(self.learn))  #appending label callback

        self.learn.model = self.learn.model.to(self._device)

        # _set_multigpu_callback(self) # MultiGPU doesn't work for U-Net. (Fastai-Forums)
        if pretrained_path is not None:
            self.load(pretrained_path)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        return [*self._resnet_family]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a Unet like classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================
        
        :returns: `UnetClassifier` Object
        """
        return cls.from_emd(data, emd_path)

    @classmethod
    def from_emd(cls, data, emd_path):
        """
        Creates a Unet like classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================
        
        :returns: `UnetClassifier` Object
        """
        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd['ModelParameters']

        try:
            class_mapping = {i['Value']: i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value']: i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue']: i['ClassName'] for i in emd['Classes']}
            color_mapping = {i['ClassValue']: i['Color'] for i in emd['Classes']}

        resize_to = emd.get('resize_to')

        if data is None:
            data = _EmptyData(path=tempfile.TemporaryDirectory().name, loss_func=None, c=len(class_mapping) + 1,
                              chip_size=emd['ImageHeight'])
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data._is_multispectral = emd.get('IsMultispectral', False)
            if data._is_multispectral:
                data._bands = emd.get('Bands')
                data._imagery_type = emd.get("ImageryType")
                data._extract_bands = emd.get("ExtractBands")
                data._train_tail = False # Hardcoded because we are never going to train a model with empty data
                normalization_stats = emd.get("NormalizationStats")
                for _stat in normalization_stats:
                    if normalization_stats[_stat] is not None:
                        normalization_stats[_stat] = torch.tensor(normalization_stats[_stat])
                    setattr(data, ('_'+_stat), normalization_stats[_stat])
                data._do_normalize = emd.get("DoNormalize")

        data.resize_to = resize_to        

        return cls(data, **model_params, pretrained_path=str(model_file))

    @property
    def _model_metrics(self):
        return {'accuracy': self._get_model_metrics()}
    
    def _create_emd(self, path):
        import random
        super()._create_emd(path)

        self._emd_template["Framework"] = "arcgis.learn.models._inferencing"
        self._emd_template["ModelConfiguration"] = "_unet"
        self._emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        self._emd_template["ExtractBands"] = [0, 1, 2]

        self._emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else \
            self._data.color_mapping[inverse_class_mapping[class_name]]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())

        self._emd_template["IsMultispectral"] = getattr(self, '_is_multispectral', False)
        if self._emd_template["IsMultispectral"]:
            self._emd_template["Bands"] = self._data._bands
            self._emd_template["ImageryType"] = self._data._imagery_type
            self._emd_template["ExtractBands"] = self._data._extract_bands
            self._emd_template["NormalizationStats"] = {
                "band_min_values": self._data._band_min_values, 
                "band_max_values": self._data._band_max_values, 
                "band_mean_values": self._data._band_mean_values, 
                "band_std_values": self._data._band_std_values, 
                "scaled_min_values": self._data._scaled_min_values, 
                "scaled_max_values": self._data._scaled_max_values, 
                "scaled_mean_values": self._data._scaled_mean_values, 
                "scaled_std_values": self._data._scaled_std_values
            }
            for _stat in self._emd_template["NormalizationStats"]:
                if self._emd_template["NormalizationStats"][_stat] is not None:
                    self._emd_template["NormalizationStats"][_stat] = self._emd_template["NormalizationStats"][_stat].tolist()
            self._emd_template["DoNormalize"] = self._data._do_normalize

        json.dump(self._emd_template, open(path.with_suffix('.emd'), 'w'), indent=4)
        return path.stem

    def _predict_batch(self, imagetensor_batch):
        predictions = self.learn.model.eval()(imagetensor_batch.to(self._device).float()).detach().cpu()
        return  predictions.max(dim=1)[1]

    #def _show_results_multispectral(self, nrows=3, index=0, type_ds='valid', rgb_bands=None, nodata=0, alpha=0.7, imsize=5, top=0.97): # Proposed Parameters 
    def _show_results_multispectral(self, rows=5, alpha=0.7, **kwargs): # parameters adjusted in kwargs
        import matplotlib.pyplot as plt
        from .._data import _tensor_scaler

        # Get Number of items
        nrows = rows
        ncols=2

        ds = self._data.valid_ds
        if kwargs.get('type_ds', None) is not None:
            type_ds = kwargs.get('type_ds')
            if hasattr(self._data, type_ds):
                ds = getattr(self._data, type_ds)
            else:
                e = Exception(f'could not find {str(type_ds)} in data.')
                raise(e)

        rgb_bands = self._data._symbology_rgb_bands
        if kwargs.get('rgb_bands', None) is not None:
            rgb_bands = kwargs.get('rgb_bands')

        nodata = 0
        if kwargs.get('nodata', None) is not None:
            nodata = kwargs.get('nodata')

        index = 0
        if kwargs.get('index', None) is not None:
            index = kwargs.get('index')

        imsize = 5
        if kwargs.get('imsize', None) is not None:
            imsize = kwargs.get('imsize')

        do_scale = True
        if kwargs.get('do_scale', None) is not None:
            do_scale = kwargs.get('do_scale')  

        do_normalize = True
        if hasattr(self._data, '_do_normalize'):
            do_normalize = getattr(self._data, '_do_normalize', True)
        if kwargs.get('do_normalize', None) is not None:
            do_normalize = kwargs.get('do_normalize')  

        title_font_size = 16
        if kwargs.get('top', None) is not None:
            top = kwargs.get('top')
        else:
            top = 1 - (math.sqrt(title_font_size)/math.sqrt(100*nrows*imsize))


        e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
        symbology_bands = []
        if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
            raise(e)
        for b in rgb_bands:
            if type(b) == str:
                b_index = self._bands.index(b)
            elif type(b) == int:
                self._bands[b] # To check if the band index specified by the user really exists.
                b_index = b
            else:
                raise(e)
            symbology_bands.append(b_index)

        # Get Batch
        x_batch, y_batch = [], []
        for i in range(index, index+nrows):
            x_batch.append(ds.x[i].data)
            y_batch.append(ds.y[i].data[0])
        x_batch = torch.stack(x_batch)

        # Scaling and normalization
        if do_scale:
            x_batch = _tensor_scaler(x_batch, self._data._band_min_values, self._data._band_max_values, mode='minmax')
        symbology_x_batch = x_batch[:, symbology_bands].cpu().numpy() # Scaled Images 0-1 for plotting
        if do_normalize:
            x_batch = ( x_batch - self._data._scaled_mean_values.view(1, -1, 1, 1) ) / self._data._scaled_std_values.view(1, -1, 1, 1)

        # Extract Bands
        if hasattr(self._data, '_extract_bands_tfm'):
            x_batch = self._data._extract_bands_tfm((x_batch, None))[0]

        # Get Predictions
        predictions = []
        for i in range(0, x_batch.shape[0], self._data.batch_size):
            predictions.append(self._predict_batch(x_batch[i:i+self._data.batch_size]))

        # Channel first to channel last for plotting
        symbology_x_batch = np.rollaxis(symbology_x_batch, 1, 4)
        if symbology_x_batch.max() < 1.5:
            symbology_x_batch = symbology_x_batch.clip(0, 1)
        y_batch = torch.stack(y_batch).cpu().numpy()
        predictions = torch.cat(predictions).cpu().numpy()

        #return x_batch, y_batch, predictions
        
        # Size for plotting
        fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
        fig.suptitle('Ground Truth / Predictions', fontsize=title_font_size)
        for r in range(nrows):
            ax[r][0].imshow(symbology_x_batch[r])
            y_rgb = _class_array_to_rbg(y_batch[r], self._data._multispectral_color_mapping, nodata)
            ax[r][0].imshow(y_rgb, alpha=alpha)
            ax[r][0].axis('off')
            ax[r][1].imshow(symbology_x_batch[r])
            p_rgb = _class_array_to_rbg(predictions[r], self._data._multispectral_color_mapping, nodata)
            ax[r][1].imshow(p_rgb, alpha=alpha)
            ax[r][1].axis('off')
            plt.subplots_adjust(top=top)

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        self.learn.callbacks = [x for x in self.learn.callbacks if not isinstance(x, LabelCallback)]
        if rows > self._data.batch_size:
            rows = self._data.batch_size
        self.learn.show_results(rows=rows, **kwargs)

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get('checkpoint', True)
        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            model_accuracy = np.min(self.learn.recorder.metrics)

        return float(model_accuracy)



class InvitationManager(object):
    """

    The `InvitationManager` provides functionality to see the existing invitations
    set out via email to your organization.  The manager has the ability to delete
    any invitation sent out by an organization.

    """
    _gis = None
    _url = None
    _invites = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        self._url = url
        self._gis = gis
    #----------------------------------------------------------------------
    def __str__(self):
        return "<InvitationManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<InvitationManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __len__(self):
        url = self._url
        params = {
            'f' : 'json',
            'num':100,
            'start':1
        }
        res = self._gis._con.get(url, params)
        return res['total']
    #----------------------------------------------------------------------
    def list(self):
        """
        Returns all the organizations invitations

        :returns: List

        """
        invites = []
        from arcgis.gis import GIS
        isinstance(self._gis, GIS)
        url = self._url
        params = {
            'f' : 'json',
            'num':100,
            'start':1
        }
        res = self._gis._con.get(url, params)
        invites = res["invitations"]
        while res['nextStart'] > -1:
            params['start'] += 100
            res = self._gis._con.get(url, params)
            invites.extend(res["invitations"])
        return invites
    def get(self, invite_id):
        """
        Returns information about a single invitation

        :returns: Dict

        """
        url = self._url + "/{id}".format(id=invite_id)
        params = {
            'f' : 'json'
        }
        return self._gis._con.get(url, params)

    #----------------------------------------------------------------------
    def delete(self, invite_id):
        """
        deletes an invitation by ID

        :returns: Boolean
        """
        url = self._url + "/{id}/delete".format(id=invite_id)
        params = {
            'f' : 'json'
        }
        res = self._gis._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res


"""
These functions help you use elevation analysis
"""

import logging as _logging
import arcgis
from datetime import datetime
from arcgis.features import FeatureSet
from arcgis.mapping import MapImageLayer
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._support import _execute_gp_tool

_log = _logging.getLogger(__name__)

_use_async = True

def profile(input_line_features: FeatureSet = {'exceededTransferLimit': False,
                                               'spatialReference': {'latestWkid': 3857, 'wkid': 102100},
                                               'geometryType': 'esriGeometryPolyline',
                                               'fields': [{'name': 'OID', 'type': 'esriFieldTypeOID', 'alias': 'OID'},
                                                          {'name': 'Shape_Length', 'type': 'esriFieldTypeDouble',
                                                           'alias': 'Shape_Length'}], 'displayFieldName': '',
                                               'features': []},
            profile_id_field: str = None,
            dem_resolution: str = None,
            maximum_sample_distance: float = None,
            maximum_sample_distance_units: str = """Meters""",
            gis=None,
            future=False) -> FeatureSet:
    """
    .. image:: _static/images/elevation_profile/elevation_profile.png

    The profile method is used to create profiles along input lines from which a profile graph can be created.

    In asynchronous mode, the maximum number of input line features that can be accepted by the task for each request is 1000.

    =====================================    ===========================================================================
    **Argument**                             **Description**
    -------------------------------------    ---------------------------------------------------------------------------
    input_line_features                      Required featureset. The line features that will be profiled over the surface.
    -------------------------------------    ---------------------------------------------------------------------------
    profile_id_field                         Optional string. A unique identifier to tie profiles to their corresponding input line features.
    -------------------------------------    ---------------------------------------------------------------------------
    dem_resolution                           Optional string. The approximate spatial resolution (cell size) of the source elevation data used for the calculation.
                                             The resolution values are an approximation of the spatial resolution of the digital elevation model. While many elevation sources are distributed in units of arc seconds, the keyword is an approximation of those resolutions in meters for easier understanding.
    -------------------------------------    ---------------------------------------------------------------------------
    maximum_sample_distance                  Optional float. The maximum sampling distance along the line to sample elevation values.
    -------------------------------------    ---------------------------------------------------------------------------
    maximum_sample_distance_units            Optional string. The units for the MaximumSampleDistance.

                                             Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    -------------------------------------    ---------------------------------------------------------------------------
    future                                   Optional boolean. If True, the result will be a `GPJob` and results will be returned asynchronously.
    =====================================    ===========================================================================

    :Returns: Output Profile as a FeatureSet

    .. code-block:: python

        USAGE EXAMPLE: To create profile of mountains feature.
        elevation = profile(input_line_features=mountain_fs,
                            dem_resolution='FINEST',
                            maximum_sample_distance=500,
                            maximum_sample_distance_units='Meters')
    """
    kwargs = locals()

    param_db = {
        "input_line_features": (FeatureSet, "InputLineFeatures"),
        "profile_id_field": (str, "ProfileIDField"),
        "dem_resolution": (str, "DEMResolution"),
        "maximum_sample_distance": (float, "MaximumSampleDistance"),
        "maximum_sample_distance_units": (str, "MaximumSampleDistanceUnits"),
        "output_profile": (FeatureSet, "Output Profile"),
    }
    return_values = [
        {"name": "output_profile", "display_name": "Output Profile", "type": FeatureSet},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.elevation.url

    return _execute_gp_tool(gis, "Profile", kwargs, param_db, return_values, _use_async, url, future=future)


def viewshed(input_points: FeatureSet = {'exceededTransferLimit': False,
                                         'spatialReference': {'latestWkid': 3857, 'wkid': 102100},
                                         'geometryType': 'esriGeometryPoint',
                                         'fields': [{'name': 'OID', 'type': 'esriFieldTypeOID', 'alias': 'OID'},
                                                    {'name': 'offseta', 'type': 'esriFieldTypeDouble',
                                                     'alias': 'offseta'},
                                                    {'name': 'offsetb', 'type': 'esriFieldTypeDouble',
                                                     'alias': 'offsetb'}], 'displayFieldName': '', 'features': []},
             maximum_distance: float = None,
             maximum_distance_units: str = """Meters""",
             dem_resolution: str = None,
             observer_height: float = None,
             observer_height_units: str = """Meters""",
             surface_offset: float = None,
             surface_offset_units: str = """Meters""",
             generalize_viewshed_polygons: bool = True,
             gis=None,
             future=False) -> FeatureSet:
    """
    .. image:: _static/images/elevation_viewshed/elevation_viewshed.png

    The ``viewshed`` method is used to identify visible areas based on observer locations you provide as well as ArcGIS Online Elevation data.

    ===============================    =========================================================
    **Parameter**                      **Description**
    -------------------------------    ---------------------------------------------------------
    input_points                       Required FeatureSet. The point features to use as the observer locations. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------    ---------------------------------------------------------
    maximum_distance                   Optional float. This is a cutoff distance where the computation of visible areas stops.
                                       Beyond this distance, it is unknown whether the analysis points and the other objects can see each other.

                                       It is useful for modeling current weather conditions or a given time of day, such as dusk. Large values increase computation time.

                                       Unless specified, a default maximum distance will be computed based on the resolution and extent of the source DEM.
                                       The allowed maximum value is 50 kilometers.

                                       Use ``maximum_distance_units`` to set the units for ``maximum_distance``.
    -------------------------------    ---------------------------------------------------------
    maximum_distance_units             Optional string. The units for the ``maximum_distance`` parameter.

                                       Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles'].

                                       The default is 'Meters'.
    -------------------------------    ---------------------------------------------------------
    dem_resolution                     Optional string. The approximate spatial resolution (cell size) of the source elevation data used for the calculation.
                                       The resolution values are an approximation of the spatial resolution of the digital elevation model.
                                       While many elevation sources are distributed in units of arc seconds, the keyword is an approximation of those resolutions in meters for easier understanding.

                                       Choice list:[' ', 'FINEST', '10m', '30m', '90m'].

                                       The default is 90m.
    -------------------------------    ---------------------------------------------------------
    observer_height                    Optional float. This is the height above the ground of the observer locations.

                                       The default is 1.75 meters, which is approximately the average height of a person.
                                       If you are looking from an elevated location, such as an observation tower or a tall building, use that height instead.

                                       Use ``observer_height_units`` to set the units for ``observer_height``.
    -------------------------------    ---------------------------------------------------------
    observer_height_units              Optional string. The units for the ``observer_height`` parameter.

                                       Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    -------------------------------    ---------------------------------------------------------
    surface_offset                     Optional float. The height above the surface of the object you are trying to see.

                                       The default value is 0.0. If you are trying to see buildings or wind turbines, use their height here.
    -------------------------------    ---------------------------------------------------------
    surface_offset_units               Optional string. The units for the ``surface_offset`` parameter.

                                       Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    -------------------------------    ---------------------------------------------------------
    generalize_viewshed_polygons       Optional boolean. Determines whether or not the viewshed polygons are to be generalized.

                                       The viewshed calculation is based on a raster elevation model that creates a result with
                                       stair-stepped edges. To create a more pleasing appearance and improve performance, the
                                       default behavior is to generalize the polygons. The generalization process smooths the
                                       boundary of the visible areas and may remove some single-cell visible areas.
    -------------------------------    ---------------------------------------------------------
    future                             Optional boolean. If True, the result will be a `GPJob` and results will be returned asynchronously.
    ===============================    =========================================================

    :returns: output_viewshed - Output Viewshed as a FeatureSet (polygons of visible areas for a given set of input observation points.)

    .. code-block:: python

        # USAGE EXAMPLE: To identify visible areas from esri headquarter office.
        visible_windfarms = viewshed(input_points=hq_fs,
                                     maximum_distance=200,
                                     maximum_distance_units='Meters',
                                     observer_height=6,
                                     observer_height_units='Feet',
                                     surface_offset=100,
                                     surface_offset_units='Meters',
                                     generalize_viewshed_polygons=True)
    """
    kwargs = locals()

    param_db = {
        "input_points": (FeatureSet, "InputPoints"),
        "maximum_distance": (float, "MaximumDistance"),
        "maximum_distance_units": (str, "MaximumDistanceUnits"),
        "dem_resolution": (str, "DEMResolution"),
        "observer_height": (float, "ObserverHeight"),
        "observer_height_units": (str, "ObserverHeightUnits"),
        "surface_offset": (float, "SurfaceOffset"),
        "surface_offset_units": (str, "SurfaceOffsetUnits"),
        "generalize_viewshed_polygons": (bool, "GeneralizeViewshedPolygons"),
        "output_viewshed": (FeatureSet, "Output Viewshed"),
    }
    return_values = [
        {"name": "output_viewshed", "display_name": "Output Viewshed", "type": FeatureSet},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.elevation.url

    return _execute_gp_tool(gis, "Viewshed", kwargs, param_db, return_values, _use_async, url, future=future)


def summarize_elevation(input_features: FeatureSet = {},
                        feature_id_field: str = None,
                        dem_resolution: str = None,
                        include_slope_aspect: bool = False,
                        gis=None,
                        future=False) -> FeatureSet:
    """
    .. image:: _static/images/summarize_elevation/summarize_elevation.png

    The ``summarize_elevation`` method calculates summary statistics for features you provide based
    on ArcGIS Online Elevation data. It accepts point, line, or polygon input and returns statistics
    for the elevation, slope, and aspect of the features.

    =========================    =========================================================
    **Parameter**                **Description**
    -------------------------    ---------------------------------------------------------
    input_features               Reqauired FeatureSet. Input features to summarize the elevation for. The features can be point, line, or area. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    feature_id_field             Optional string. The unique ID field to use for the input features.
    -------------------------    ---------------------------------------------------------
    dem_resolution               Optional string. The approximate spatial resolution (cell size) of the source elevation data used for the calculation.

                                 Choice list:[' ', 'FINEST', '10m', '30m', '90m']

                                 The default value is None.
    -------------------------    ---------------------------------------------------------
    include_slope_aspect         Optional boolean. Determines if slope and aspect for the input feature(s) will be included in the output. The slope and aspect values in the output are in degrees.

                                 The default value is False.
    -------------------------    ---------------------------------------------------------
    future                       Optional boolean. If True, the result will be a `GPJob` and results will be returned asynchronously.
    =========================    =========================================================

    :returns: result_layer : Output Summary as a FeatureSet

    .. code-block:: python

        # USAGE EXAMPLE: To calculate summary statistics for mountain polyline features.
        summarize = summarize_elevation(input_features=mountain_fs,
                           dem_resolution='FINEST',
                           include_slope_aspect=True)
    """
    kwargs = locals()

    param_db = {
        "input_features": (FeatureSet, "InputFeatures"),
        "feature_id_field": (str, "FeatureIDField"),
        "dem_resolution": (str, "DEMResolution"),
        "include_slope_aspect": (bool, "IncludeSlopeAspect"),
        "output_summary": (FeatureSet, "Output Summary"),
    }
    return_values = [
        {"name": "output_summary", "display_name": "Output Summary", "type": FeatureSet},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.elevation.url

    return _execute_gp_tool(gis, "SummarizeElevation", kwargs, param_db, return_values, _use_async, url, future=future)





from ._ssd import SingleShotDetector
from ._inferencing import _DynamicSSD
from ._unet import UnetClassifier
from ._retinanet import RetinaNet
from ._ner import EntityRecognizer
from ._classifier import FeatureClassifier
from ._pspnet import PSPNetClassifier
from ._maskrcnn import MaskRCNN


import os
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap

########################################################################
class SecurityManager(object):
    """
    This resource is a container for all resources and operations
    pertaining to security in your ArcGIS Notebook Server site.
    """
    _url = None
    _gis = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<SecurityManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    @property
    def configuration(self):
        """
        This resource returns the currently active security configuration
        of your ArcGIS Notebook Server site. A security configuration
        involves the following pieces of information, and can be modified
        using the Update operation.

        **Configuration Properties**

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        httpsProtocols	       The TLS protocols ArcGIS Notebook Server will use. Values must be comma-separated.
        ------------------     --------------------------------------------------------------------
        cipherSuites	       The cipher suites ArcGIS Notebook Server will use. Values must be comma-separated.
        ------------------     --------------------------------------------------------------------
        authenticationMode     Specifies the authentication mode used by ArcGIS Notebook Server. When ArcGIS Notebook Server is federated with an Enterprise portal, this property can be included and set to ARCGIS_PORTAL_TOKEN
                               Values: ARCGIS_TOKEN | ARCGIS_PORTAL_TOKEN
        ------------------     --------------------------------------------------------------------
        authenticationTier     The tier at which requests to access GIS services will be authenticated. You should not use this directory to modify your setting. The default value when your site is first created is NOTEBOOK_SERVER, but when you federate the site with your portal, the value is changed to ARCGIS_PORTAL.
        ------------------     --------------------------------------------------------------------
        tokenServiceKey	       The key used to encrypt tokens.
        ==================     ====================================================================


        **Portal Properties**

        ==========================     ====================================================================
        **Argument**                   **Description**
        --------------------------     --------------------------------------------------------------------
        portalMode	               Must be the value ARCGIS_PORTAL_FEDERATION.
        --------------------------     --------------------------------------------------------------------
        portalSecretKey	               The key obtained after federating ArcGIS Notebook Server with portal.
        --------------------------     --------------------------------------------------------------------
        portalURL	               The URL of your portal, in the format https://webadaptorhost.domain.com/webadaptorname.
        --------------------------     --------------------------------------------------------------------
        referer	                       The referer specified when generating the token.
        --------------------------     --------------------------------------------------------------------
        serverId	               The ID of the server federated with the portal.
        --------------------------     --------------------------------------------------------------------
        serverUrl	               The external URL of the federated ArcGIS Notebook Server, in the format https://webadaptorhost.domain.com/webadaptorname.
        --------------------------     --------------------------------------------------------------------
        token                          A token obtained from the portal for initial validation of the ArcGIS Notebook Server.
        --------------------------     --------------------------------------------------------------------
        webgisServerTrustKey           A key for establishing trust between servers that are federated with the same portal.
        --------------------------     --------------------------------------------------------------------
        privateHostingServerUrl	       The private URL of the portal's hosting server.
        --------------------------     --------------------------------------------------------------------
        privatePortalUrl               The private URL of the portal.
        ==========================     ====================================================================



        When Setting the `configuration` the following must be defined and provided as a dictionary.

        ==================     ====================================================================
        **Key**                **Value**
        ------------------     --------------------------------------------------------------------
        portalProperties       Required Dict. Portal properties represented as a JSON object.
        ------------------     --------------------------------------------------------------------
        httpsProtocols         Required String. The TLS protocols ArcGIS Server will use. Valid options are TLSv1, TLSv1.1, and TLSv1.2; values must be comma separated. By default, only TLSv1.2 is enabled.
        ------------------     --------------------------------------------------------------------
        cipherSuites	       Required String. The cipher suites ArcGIS Notebook Server will use. Valid options are: TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256, TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256, TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA, TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA, TLS_RSA_WITH_AES_128_GCM_SHA256, TLS_RSA_WITH_AES_128_CBC_SHA256, TLS_RSA_WITH_AES_128_CBC_SHA, and TLS_RSA_WITH_3DES_EDE_CBC_SHA.
                               By default, all of the above options are enabled. Values must be comma separated
        ------------------     --------------------------------------------------------------------
        authenticationTier     Required String. The tier at which requests to access GIS services will be authenticated. You should only indicate your setting here, rather than using this operation to modify your setting. The default value when your site is first created is NOTEBOOK_SERVER, but when you federate the site with your portal, the value is changed to ARCGIS_PORTAL. Values: WEB_ADAPTOR | NOTEBOOK_SERVER | ARCGIS_PORTAL
        ==================     ====================================================================

        """
        url = self._url + "/config"
        params = {'f' : 'json'}
        return self._con.get(url, params)
    #----------------------------------------------------------------------
    @configuration.setter
    def configuration(self, settings):
        """
        This resource returns the currently active security configuration
        of your ArcGIS Notebook Server site. A security configuration
        involves the following pieces of information, and can be modified
        using the Update operation.

        **Configuration Properties**

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        httpsProtocols	       The TLS protocols ArcGIS Notebook Server will use. Values must be comma-separated.
        ------------------     --------------------------------------------------------------------
        cipherSuites	       The cipher suites ArcGIS Notebook Server will use. Values must be comma-separated.
        ------------------     --------------------------------------------------------------------
        authenticationMode     Specifies the authentication mode used by ArcGIS Notebook Server. When ArcGIS Notebook Server is federated with an Enterprise portal, this property can be included and set to ARCGIS_PORTAL_TOKEN
                               Values: ARCGIS_TOKEN | ARCGIS_PORTAL_TOKEN
        ------------------     --------------------------------------------------------------------
        authenticationTier     The tier at which requests to access GIS services will be authenticated. You should not use this directory to modify your setting. The default value when your site is first created is NOTEBOOK_SERVER, but when you federate the site with your portal, the value is changed to ARCGIS_PORTAL.
        ------------------     --------------------------------------------------------------------
        tokenServiceKey	       The key used to encrypt tokens.
        ==================     ====================================================================


        **Portal Properties**

        ==========================     ====================================================================
        **Argument**                   **Description**
        --------------------------     --------------------------------------------------------------------
        portalMode	               Must be the value ARCGIS_PORTAL_FEDERATION.
        --------------------------     --------------------------------------------------------------------
        portalSecretKey	               The key obtained after federating ArcGIS Notebook Server with portal.
        --------------------------     --------------------------------------------------------------------
        portalURL	               The URL of your portal, in the format https://webadaptorhost.domain.com/webadaptorname.
        --------------------------     --------------------------------------------------------------------
        referer	                       The referer specified when generating the token.
        --------------------------     --------------------------------------------------------------------
        serverId	               The ID of the server federated with the portal.
        --------------------------     --------------------------------------------------------------------
        serverUrl	               The external URL of the federated ArcGIS Notebook Server, in the format https://webadaptorhost.domain.com/webadaptorname.
        --------------------------     --------------------------------------------------------------------
        token                          A token obtained from the portal for initial validation of the ArcGIS Notebook Server.
        --------------------------     --------------------------------------------------------------------
        webgisServerTrustKey           A key for establishing trust between servers that are federated with the same portal.
        --------------------------     --------------------------------------------------------------------
        privateHostingServerUrl	       The private URL of the portal's hosting server.
        --------------------------     --------------------------------------------------------------------
        privatePortalUrl               The private URL of the portal.
        ==========================     ====================================================================



        When Setting the `configuration` the following must be defined and provided as a dictionary.

        ==================     ====================================================================
        **Key**                **Value**
        ------------------     --------------------------------------------------------------------
        portalProperties       Required Dict. Portal properties represented as a JSON object.
        ------------------     --------------------------------------------------------------------
        httpsProtocols         Required String. The TLS protocols ArcGIS Server will use. Valid options are TLSv1, TLSv1.1, and TLSv1.2; values must be comma separated. By default, only TLSv1.2 is enabled.
        ------------------     --------------------------------------------------------------------
        cipherSuites	       Required String. The cipher suites ArcGIS Notebook Server will use. Valid options are: TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256, TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256, TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA, TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA, TLS_RSA_WITH_AES_128_GCM_SHA256, TLS_RSA_WITH_AES_128_CBC_SHA256, TLS_RSA_WITH_AES_128_CBC_SHA, and TLS_RSA_WITH_3DES_EDE_CBC_SHA.
                               By default, all of the above options are enabled. Values must be comma separated
        ------------------     --------------------------------------------------------------------
        authenticationTier     Required String. The tier at which requests to access GIS services will be authenticated. You should only indicate your setting here, rather than using this operation to modify your setting. The default value when your site is first created is NOTEBOOK_SERVER, but when you federate the site with your portal, the value is changed to ARCGIS_PORTAL. Values: WEB_ADAPTOR | NOTEBOOK_SERVER | ARCGIS_PORTAL
        ==================     ====================================================================

        """
        url = self._url + "/config/update"
        params = {'f' : 'json'}
        current = dict(self.configuration)
        for k, v in settings.items():
            if k in current:
                params[k] = settings[k]
                current.pop(k)
        params.update(current)
        return self._con.post(url, params)



import os, sys
from contextlib import ExitStack, redirect_stdout, redirect_stderr, suppress
from io import StringIO
def run_and_hide(fn, **kwargs):
    """runs and hides the code's output"""
    err = None
    res = None
    with ExitStack() as stack, suppress(Exception) as s3, suppress(ValueError) as s2, suppress(SystemError) as s1:
        null_stream = StringIO()
        with redirect_stdout(null_stream) as s4:
            try:
                old = sys.tracebacklimit
                sys.tracebacklimit = 0
            except:
                sys.tracebacklimit = 0
                old = 1000
            stack.enter_context(redirect_stdout(null_stream))
            try:
                res = fn(**kwargs)
            except Exception as errmsg:
                err = errmsg
            finally:
                stack.close()
                sys.tracebacklimit = 1000
    if err:
        raise err
    return res


"""
These tools help answer one of the most common questions posed in spatial analysis: What is near what?

create_buffers() creates areas of a specified distance from features.
"""
import json as _json

import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet, FeatureCollection
from arcgis.geoprocessing._support import _execute_gp_tool
from ._util import _id_generator, _feature_input, _set_context, _create_output_service, GAJob

_log = _logging.getLogger(__name__)

_use_async = True

def create_buffers(input_layer,
                   distance=1,
                   distance_unit="Miles",
                   field=None,
                   method="Planar",
                   dissolve_option="None",
                   dissolve_fields=None,
                   summary_fields=None,
                   multipart=False,
                   output_name=None,
                   context=None,
                   gis=None,
                   future=False):
    """

    .. image:: _static/images/create_buffers_geo/create_buffers_geo.png

    Buffers are typically used to create areas that can be further analyzed 
    using other tools such as ``aggregate_points``. For example, ask the question, 
    "What buildings are within one mile of the school?" The answer can be found 
    by creating a one-mile buffer around the school and overlaying the buffer 
    with the layer containing building footprints. The end result is a layer 
    of those buildings within one mile of the school.

    ================================================    =========================================================
    **Parameter**                                       **Description**
    ------------------------------------------------    ---------------------------------------------------------
    input_layer                                         Required layer. The point, line, or polygon features to be buffered. 
                                                        See :ref:`Feature Input<FeatureInput>`.
    ------------------------------------------------    ---------------------------------------------------------
    distance (Required if field is not provided)        Optional float. A float value used to buffer the input features. 
                                                        You must supply a value for either the distance or field parameter. 
                                                        You can only enter a single distance value. The units of the 
                                                        distance value are supplied by the ``distance_unit`` parameter.

                                                        The default value is 1.
    ------------------------------------------------    ---------------------------------------------------------
    distance_unit (Required if distance is used)        Optional string. The linear unit to be used with the value specified in distance.  

                                                        Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']    

                                                        The default value is "Miles"
    ------------------------------------------------    ---------------------------------------------------------
    field (Required if distance not provided)           Optional string. A field on the ``input_layer`` containing a buffer distance or a field expression.
                                                        A buffer expression must begin with an equal sign (=). To learn more about buffer expressions 
                                                        see: `Buffer Expressions <https://developers.arcgis.com/rest/services-reference/bufferexpressions.htm>`_
    ------------------------------------------------    ---------------------------------------------------------
    method                                              Optional string. The method used to apply the buffer with. There are two methods to choose from:
    
                                                        Choice list:['Geodesic', 'Planar']

                                                        * ``Planar`` - This method applies a Euclidean buffers and is appropriate for local analysis on projected data. This is the default.
                                                        * ``Geodesic`` - This method is appropriate for large areas and any geographic coordinate system.
    ------------------------------------------------    ---------------------------------------------------------
    dissolve_option                                     Optional string. Determines how output polygon attributes are processed.
    
                                                        Choice list:['All', 'List', 'None']    

                                                        +----------------------------------+---------------------------------------------------------------------------------------------------+
                                                        |Value                             | Description                                                                                       |
                                                        +----------------------------------+---------------------------------------------------------------------------------------------------+
                                                        | All - All features are dissolved | You can calculate summary statistics and determine if you want multipart or single part features. |
                                                        | into one feature.                |                                                                                                   |
                                                        +----------------------------------+---------------------------------------------------------------------------------------------------+
                                                        | List - Features with the same    | You can calculate summary statistics and determine if you want multipart or single part features. |
                                                        | value in the specified field     |                                                                                                   |
                                                        | will be dissolve together.       |                                                                                                   |
                                                        +----------------------------------+---------------------------------------------------------------------------------------------------+
                                                        | None - No features are dissolved.| There are no additional dissolve options.                                                         |
                                                        +----------------------------------+---------------------------------------------------------------------------------------------------+
    ------------------------------------------------    ---------------------------------------------------------
    dissolve_field                                      Specifies the fields to dissolve on. Multiple fields may be provided.
    ------------------------------------------------    ---------------------------------------------------------
    summary_fields                                      Optional string. A list of field names and statistical summary types 
                                                        that you want to calculate for resulting polygons. Summary statistics 
                                                        are only available if dissolveOption = List or All. By default, all 
                                                        statistics are returned.

                                                        Example: [{"statisticType": "statistic type", "onStatisticField": "field name"}, ..}]
            
                                                        fieldName is the name of the fields in the input point layer.

                                                        statisticType is one of the following for numeric fields:

                                                            * ``Count`` - Totals the number of values of all the points in each polygon.
                                                            * ``Sum`` - Adds the total value of all the points in each polygon.
                                                            * ``Mean`` - Calculates the average of all the points in each polygon.
                                                            * ``Min`` - Finds the smallest value of all the points in each polygon.
                                                            * ``Max`` - Finds the largest value of all the points in each polygon.
                                                            * ``Range`` - Finds the difference between the Min and Max values.
                                                            * ``Stddev`` - Finds the standard deviation of all the points in each polygon.
                                                            * ``Var`` - Finds the variance of all the points in each polygon.

                                                        statisticType is the following for string fields:

                                                            * ``Count`` - Totals the number of strings for all the points in each polygon.
                                                            * ``Any`` - Returns a sample string of a point in each polygon.

    ------------------------------------------------    ---------------------------------------------------------
    multipart                                           Optional boolean. Determines if output features are multipart or single part. 
                                                        This option is only available if a ``dissolve_option`` is applied.
    ------------------------------------------------    ---------------------------------------------------------
    output_name                                         Optional string. The task will create a feature service of the results. You define the name of the service.
    ------------------------------------------------    ---------------------------------------------------------
    gis                                                 Optional, the GIS on which this tool runs. If not specified, the active GIS is used.  
    ------------------------------------------------    ---------------------------------------------------------
    context                                             Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                        #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                        #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                                        #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                        #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ------------------------------------------------    ---------------------------------------------------------
    future                                              Optional boolean. If 'True', the value is returned as a GPJob.

                                                        The default value is 'False'
    ================================================    =========================================================

    :returns: Output Features as a feature layer collection item
    
    .. code-block:: python

            # Usage Example: To create buffer based on distance field.

            buffer = create_buffers(input_layer=lyr,
                                    field='dist',
                                    method='Geodesic',
                                    dissolve_option='All',
                                    dissolve_fields='Date')    
    """
    kwargs = locals()


    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    if isinstance(input_layer, FeatureCollection) and \
       'layers' in input_layer.properties and \
       len(input_layer.properties.layers) > 0:
        input_layer = _FeatureSet.from_dict(
            featureset_dict=input_layer._lazy_properties.layers[0].featureSet)

    params = {}
    for key, value in kwargs.items():
        if key != 'field':
            if value is not None:
                params[key] = value
        else:
            params['distance'] = None
            params['distance_unit'] = None

    if output_name is None:
        output_service_name = 'Create Buffers Analysis_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Create Buffers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)
        
    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "distance": (float, "distance"),
        "distance_unit": (str, "distanceUnit"),
        "field": (str, "field"),
        "method": (str, "method"),
        "dissolve_option": (str, "dissolveOption"),
        "dissolve_fields": (str, "dissolveFields"),
        "summary_fields": (str, "summaryFields"),
        "multipart": (bool, "multipart"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, "CreateBuffers", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "CreateBuffers", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

create_buffers.__annotations__ = {
    'distance': float,
    'distance_unit': str,
    'field': str,
    'method': str,
    'dissolve_option': str,
    'dissolve_fields': str,
    'summary_fields': str,
    'multipart': bool,
    'output_name': str,
    'context': str}


import numpy as np

HAS_OPENCV = True
try:
    import cv2
except:
    HAS_OPENCV = False


def _pad_image(image, stride):

    img_h = image.shape[0] + stride*2
    img_w = image.shape[1] + stride*2

    tmp = np.zeros((img_h, img_w, image.shape[2]))
    tmp[stride:stride+image.shape[0], stride:stride+image.shape[1], :] = image[:, :, :]
    img = tmp.astype(np.uint8)

    return img

def _get_image_chips(image, chip_dim):
    img_h, img_w, _ = image.shape
    chips_data = []
    stride = chip_dim // 2

    # Add zero-padding to enable prediction on all parts of the image
    padded_image = _pad_image(image, stride)
    img_h, img_w, _ = padded_image.shape

    start_x = 0
    start_y = 0

    if chip_dim >= img_w or chip_dim >= img_h:
        return [{'height': img_h, 'width': img_w, 'chip': image, 'xmin': start_x, 'ymin': start_y, 'predictions': []}]

    while start_x + chip_dim <= img_w:
        start_y = 0
        while start_y + chip_dim <= img_h:
            chip = padded_image[start_y: start_y + chip_dim, start_x: start_x + chip_dim, :]

            if (chip.shape[0] != chip_dim) or (chip.shape[1] != chip_dim):
                tmp = np.zeros((chip_dim, chip_dim, 3))
                tmp[0:chip.shape[0], 0:chip.shape[1], :] = chip[:, :, :]
                chip = tmp.astype(np.uint8)

            # To translate the bbox coordinates according to unpadded image
            xmin = start_x - stride
            ymin = start_y - stride

            chips_data.append({'height': chip_dim, 'width': chip_dim, 'chip': chip, 'xmin': xmin, 'ymin': ymin, 'predictions': []})
            start_y = start_y + stride

        start_x = start_x + stride

    return chips_data


def _get_transformed_predictions(chips_data):
    predictions = []
    labels = []
    scores = []
    for chip_data in chips_data:

        for prediction in chip_data['predictions']:
            prediction['xmin'] = prediction['xmin'] + chip_data['xmin']
            prediction['ymin'] = prediction['ymin'] + chip_data['ymin']

            predictions.append([
                prediction['xmin'],
                prediction['ymin'],
                prediction['width'],
                prediction['height']
            ])
            labels.append(prediction['label'].obj)
            scores.append(prediction['score'])

    return predictions, labels, scores


def _draw_predictions(frame, predictions, labels, scores=None, show_scores=True, show_labels=True, color=(255, 255, 255), fontface=0, thickness=2):
    for index, data in enumerate(predictions):
        frame = cv2.rectangle(
            frame,
            (int(data[0]), int(data[1])), (int(data[0] + data[2]), int(data[1] + data[3])),
            color,
            thickness
        )

        text_to_display = None
        if show_labels:
            text_to_display = str(labels[index])

        if show_scores and scores is not None:
            if text_to_display:
                scores[index] = float("{0:.2f}".format(scores[index]))
                text_to_display = text_to_display + ": " + str(scores[index])
            else:
                text_to_display = str(scores[index])

        if text_to_display:
            frame = cv2.putText(
                frame,
                text_to_display,
                (int(data[0]), int(data[1]) - 10),
                fontface,
                0.7,
                color,
                thickness
            )

    return frame


def _exclude_detection(data, chip_width, chip_height):
    if chip_height < chip_width:
        padding = chip_height // 4
    else:
        padding = chip_width // 4

    center_coord_x = data[0] + data[2]/2
    center_coord_y = data[1] + data[3]/2

    if center_coord_x < padding or center_coord_y < padding\
            or center_coord_x > (chip_width - padding)\
            or center_coord_y > (chip_height - padding):
        return True

    return False


"""

Contains an internal GP method used to calculate Credit Usage for Tool
This is a private method and could change without warning. Do not use.

"""
import json
import arcgis as _arcgis
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._support import _execute_gp_tool

def _estimate_credits(task, parameters, gis=None):
    """
    Estimates the number of credits a spatial analysis operation will take.

    =======================     ====================================================================
    **Argument**                **Description**
    -----------------------     --------------------------------------------------------------------
    task                        Required String. The name of the analysis tool.
    -----------------------     --------------------------------------------------------------------
    parameters                  Required String.  The input parameters for the tool.
    -----------------------     --------------------------------------------------------------------
    gis                         Optional GIS.  The enterprise connection object.
    =======================     ====================================================================

    :returns: float

    """

    if gis is None and \
       _arcgis.env.active_gis:
        gis = _arcgis.env.active_gis
    elif gis is None and \
         _arcgis.env.active_gis is None:
        raise Exception("A GIS must be provided and/or set as active.")
    if gis.version >= [6,4] and \
       gis._portal.is_arcgisonline:
        url = gis.properties['helperServices']['creditEstimation']['url']
        gptask = "EstimateCredits"
        url = "{base}/{gptask}/execute".format(base=url, gptask=gptask)
        params = {
            'f' : 'json',
            'taskName' : task,
            'taskParameters' : json.dumps(parameters)
        }
        kwargs = locals()
        param_db = {
            "task": (str, "taskName"),
            "parameters": (str, "taskParameters"),
            "credit_estimate": (str, "creditEstimate"),
        }
        return_values = [
            {"name":"credit_estimate", "display_name":"creditEstimate", "type":str},
        ]
        res = _execute_gp_tool(gis, gptask, kwargs, param_db, return_values, False, url, webtool=True, add_token=False)
        if 'cost' in res:
            return float(res['cost'])
        return res
    return



# MIT License

# Copyright (c) 2019 Hengshuang Zhao

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Based on https://github.com/hszhao/semseg

import torch
import warnings
import PIL
import numpy as np

from pdb import set_trace
import torch.nn.functional as F
import torch.nn as nn
import torch
from torchvision import models

import math
from fastai.callbacks.hooks import hook_output
from fastai.vision.learner import create_body
from fastai.callbacks.hooks import model_sizes

def initialize_weights(*models):
    for model in models:
        for module in model.modules():
            if isinstance(module, nn.Conv2d) or isinstance(module, nn.Linear):
                nn.init.kaiming_normal_(module.weight)
                if module.bias is not None:
                    module.bias.data.zero_()
            elif isinstance(module, nn.BatchNorm2d):
                module.weight.data.fill_(1)
                module.bias.data.zero_()


class _PyramidPoolingModule(nn.Module):
    """
    Creates the pyramid pooling module as in https://arxiv.org/abs/1612.01105
    Takes a feature map from the backbone and pools it at different scales
    according to the given pyramid sizes and upsamples it to original feature
    map size and concatenates it with the feature map. 
    Code from https://github.com/hszhao/semseg.
    """
    def __init__(self, in_dim, reduction_dim, setting):
        super(_PyramidPoolingModule, self).__init__()
        self.features = []

        ## Creating modules for different pyramid sizes
        for s in setting:
            self.features.append(nn.Sequential(
                nn.AdaptiveAvgPool2d(s),
                nn.Conv2d(in_dim, reduction_dim, kernel_size=1, bias=False),
                nn.BatchNorm2d(reduction_dim, momentum=.95),
                nn.ReLU(inplace=True)
            ))
        self.features = nn.ModuleList(self.features)

    def forward(self, x):
        x_size = x.size()
        out = [x]
        for f in self.features:
            ## Pass through the module which reduces its spatial size and then upsamples it.
            out.append(F.interpolate(f(x), x_size[2:], mode='bilinear', align_corners=True))
        out = torch.cat(out, 1)
        return out


def _pspnet_unet(num_classes, backbone_fn, chip_size=224, pyramid_sizes=(1, 2, 3, 6), pretrained=True):
    """
    Function which returns PPM module attached to backbone which is then used to form the Unet.
    """  
    backbone = create_body(backbone_fn, pretrained=pretrained)
    
    backbone_name = backbone_fn.__name__

    ## Support for different backbones
    if "densenet" in backbone_name or "vgg" in backbone_name:
        hookable_modules = list(backbone.children())[0]
    else:
        hookable_modules = list(backbone.children())
    
    if "vgg" in backbone_name:
        modify_dilation_index = -5
    else:
        modify_dilation_index = -2
        
    if backbone_name == 'resnet18' or backbone_name == 'resnet34':
        module_to_check = 'conv' 
    else:
        module_to_check = 'conv2'
    
    custom_idx = 0
    for i, module in enumerate(hookable_modules[modify_dilation_index:]): 
        dilation = 2 * (i + 1)
        padding = 2 * (i + 1)
        # padding = 1
        for n, m in module.named_modules():
            if module_to_check in n:
                m.dilation, m.padding, m.stride = (dilation, dilation), (padding, padding), (1, 1)
            elif 'downsample.0' in n:
                m.stride = (1, 1)                    
                
        if "vgg" in backbone_fn.__name__:
            if isinstance(module, nn.Conv2d):
                dilation = 2 * (custom_idx + 1)
                padding = 2 * (custom_idx + 1)
                module.dilation, module.padding, module.stride = (dilation, dilation), (padding, padding), (1, 1)
                custom_idx += 1
    
    ## returns the size of various activations
    feature_sizes = model_sizes(backbone, size=(chip_size, chip_size))

    ## Get number of channels in the last layer
    num_channels = feature_sizes[-1][1]

    penultimate_channels = num_channels / len(pyramid_sizes)
    ppm = _PyramidPoolingModule(num_channels, int(penultimate_channels), pyramid_sizes)

    in_final = int(penultimate_channels) * len(pyramid_sizes) + num_channels

    # Reduce channel size after pyramid pooling module to avoid CUDA OOM error.
    final_conv = nn.Conv2d(in_channels=in_final, out_channels=512, kernel_size=3, padding=1)

    ## To make Dynamic Unet work as it expects a backbone which can be indexed.
    if "densenet" in backbone_name or "vgg" in backbone_name:
        backbone = backbone[0]
    layers = [*backbone, ppm, final_conv]
    return nn.Sequential(*layers)



class PSPNet(nn.Module):
    def __init__(self, num_classes, backbone_fn, chip_size=224, pyramid_sizes=(1, 2, 3, 6), pretrained=True):
        super(PSPNet, self).__init__()        
        
        self.backbone = create_body(backbone_fn, pretrained=pretrained)
        
        backbone_name = backbone_fn.__name__

        ## Support for different backbones
        if "densenet" in backbone_name or "vgg" in backbone_name:
            hookable_modules = list(self.backbone.children())[0]
        else:
            hookable_modules = list(self.backbone.children())
        
        if "vgg" in backbone_name:
            modify_dilation_index = -5
        else:
            modify_dilation_index = -2
            
        if backbone_name == 'resnet18' or backbone_name == 'resnet34':
            module_to_check = 'conv' 
        else:
            module_to_check = 'conv2'
        
        ## Hook at the index where we need to get the auxillary logits out
        self.hook = hook_output(hookable_modules[modify_dilation_index])
        
        custom_idx = 0
        for i, module in enumerate(hookable_modules[modify_dilation_index:]): 
            dilation = 2 * (i + 1)
            padding = 2 * (i + 1)
            for n, m in module.named_modules():
                if module_to_check in n:
                    m.dilation, m.padding, m.stride = (dilation, dilation), (padding, padding), (1, 1)
                elif 'downsample.0' in n:
                    m.stride = (1, 1)                    
                    
            if "vgg" in backbone_fn.__name__:
                if isinstance(module, nn.Conv2d):
                    dilation = 2 * (custom_idx + 1)
                    padding = 2 * (custom_idx + 1)
                    module.dilation, module.padding, module.stride = (dilation, dilation), (padding, padding), (1, 1)
                    custom_idx += 1
        
        ## returns the size of various activations
        feature_sizes = model_sizes(self.backbone, size=(chip_size, chip_size))

        ## Geting the stored parameters inside of the hook
        aux_in_channels = self.hook.stored.shape[1]

        ## Get number of channels in the last layer
        num_channels = feature_sizes[-1][1]

        penultimate_channels = num_channels / len(pyramid_sizes)
        self.ppm = _PyramidPoolingModule(num_channels, int(penultimate_channels), pyramid_sizes)
        
        
        self.final = nn.Sequential(
            ## To handle case when the length of pyramid_sizes is odd
            nn.Conv2d(int(penultimate_channels) * len(pyramid_sizes) + num_channels, math.ceil(penultimate_channels), kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(math.ceil(penultimate_channels)),
            nn.ReLU(inplace=True),
            nn.Dropout(0.1),
            nn.Conv2d(math.ceil(penultimate_channels), num_classes, kernel_size=1)
        )
        
        
        self.aux_logits = nn.Conv2d(aux_in_channels, num_classes, kernel_size=1)
        
        initialize_weights(self.aux_logits)
        initialize_weights(self.ppm, self.final)

    def forward(self, x):
        x_size = x.size()
        x = self.backbone(x)
        if self.training:
            aux_l = self.aux_logits(self.hook.stored)
        
        ## Remove hook to free up memory.
        self.hook.remove()
        x = self.ppm(x)
        x = self.final(x)
        if self.training:
            return F.interpolate(x, x_size[2:], mode='bilinear', align_corners=True), F.interpolate(aux_l, x_size[2:], mode='bilinear', align_corners=True)
        else:
            return F.interpolate(x, x_size[2:], mode='bilinear', align_corners=True) 


from arcgis.gis import GIS
from arcgis.features import FeatureLayer
from arcgis._impl.common._mixins import PropertyMap
import collections
import json

def _lazy_property(fn):
    '''Decorator that makes a property lazy-evaluated.
    '''
    # http://stevenloria.com/lazy-evaluated-properties-in-python/
    attr_name = '_lazy_' + fn.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazy_property

class Hub(object):
    """
    Entry point into the Hub module. Lets you access an individual hub and its components.
       

    ================    ===============================================================
    **Argument**        **Description**
    ----------------    ---------------------------------------------------------------
    url                 Required string. If no URL is provided by user while connecting 
                        to the GIS, then the URL will be ArcGIS Online.
    ----------------    ---------------------------------------------------------------
    username            Optional string as entered while connecting to GIS. The login user name 
                        (case-sensitive).
    ----------------    ---------------------------------------------------------------
    password            Optional string as entered while connecting to GIS. If a username is 
                        provided, a password is expected.  This is case-sensitive. If the password 
                        is not provided, the user is prompted in the interactive dialog.
    ================    ===============================================================

    """
    
    def __init__(self, gis, username=None, password=None):
        self.gis = gis
        self._username = username
        self._password = password
        try:
            self._gis_id = self.gis.properties.id
        except AttributeError:
            self._gis_id = None
            
    @property
    def enterprise_org_id(self):
        """
        Returns the AGOL org id of the Enterprise Organization associated with this Hub.
        """
        try:
            self.gis.properties.portalProperties.hub
            try:
                return self.gis.properties.portalProperties.hub.settings.enterpriseOrg.orgId
            except AttributeError: 
                return  self._gis_id
        except:
            print("Hub does not exist or is inaccessible.")
            raise
                        
    @property
    def community_org_id(self):
        """
        Returns the AGOL org id of the Community Organization associated with this Hub.
        """
        try:
            self.gis.properties.portalProperties.hub
            try:
                return self.gis.properties.portalProperties.hub.settings.communityOrg.orgId
            except AttributeError:
                return  self._gis_id
        except:
            print("Hub does not exist or is inaccessible.")
            raise  
  
    @property
    def enterprise_org_url(self):
        """
        Returns the AGOL org url of the Enterprise Organization associated with this Hub.
        """
        try:
            self.gis.properties.portalProperties.hub
            try:
                self.gis.properties.portalProperties.hub.settings.enterpriseOrg
                try:
                    _url = self.gis.properties.publicSubscriptionInfo.companionOrganizations[0]['organizationUrl']
                except:
                    _url = self.gis.properties.subscriptionInfo.companionOrganizations[0]['organizationUrl']
                return "https://"+_url
            except AttributeError: 
                return self.gis.url
        except AttributeError:
            print("Hub does not exist or is inaccessible.")
            raise
        
    @property
    def community_org_url(self):
        """
        Returns the AGOL org id of the Community Organization associated with this Hub.
        """
        try:
            self.gis.properties.portalProperties.hub
            try:
                self.gis.properties.portalProperties.hub.settings.communityOrg
                try:
                    _url = self.gis.properties.publicSubscriptionInfo.companionOrganizations[0]['organizationUrl']
                except:
                    _url = self.gis.properties.subscriptionInfo.companionOrganizations[0]['organizationUrl']
                return "https://"+_url
            except AttributeError: 
                return self.gis.url
        except:
            print("Hub does not exist or is inaccessible.")
            raise
    
    @_lazy_property
    def initiatives(self):
        """
        The resource manager for Hub initiatives. See :class:`~arcgis.apps.hub.InitiativeManager`.
        """
        return InitiativeManager(self)
    
    @_lazy_property
    def events(self):
        """
        The resource manager for Hub events. See :class:`~arcgis.apps.hub.EventManager`.
        """
        return EventManager(self)
    
class Initiative(collections.OrderedDict):
    """
    Represents an initiative within a Hub. An Initiative supports 
    policy- or activity-oriented goals through workflows, tools and team collaboration.
    """
    
    def __init__(self, gis, initiativeItem):
        """
        Constructs an empty Initiative object
        """
        self.item = initiativeItem
        self._gis = gis
        try:
            self._initiativedict = self.item.get_data()
            pmap = PropertyMap(self._initiativedict)
            self.definition = pmap
        except:
            self.definition = None
            
    def __repr__(self):
        return '<%s title:"%s" owner:%s>' % (type(self).__name__, self.title, self.owner)
       
    @property
    def itemid(self):
        """
        Returns the item id of the initiative item
        """
        return self.item.id
    
    @property
    def title(self):
        """
        Returns the title of the initiative item
        """
        return self.item.title
    
    @property
    def description(self):
        """
        Getter/Setter for the initiative description
        """
        return self.item.description
    
    @description.setter
    def description(self, value):
        self.item.description = value
    
    @property
    def snippet(self):
        """
        Getter/Setter for the initiative snippet
        """
        return self.item.snippet
    
    @snippet.setter
    def snippet(self, value):
        self.item.snippet = value
    
    @property
    def owner(self):
        """
        Returns the owner of the initiative item
        """
        return self.item.owner

    @property
    def tags(self):
        """
        Returns the tags of the initiative item
        """
        return self.item.tags
    
    @property
    def url(self):
        """
        Returns the url of the initiative editor
        """
        return self.item.properties['url']
    
    @property
    def site_url(self):
        """
        Returns the url of the initiative site
        """
        return self.item.url
    
    @_lazy_property
    def indicators(self):
        """
        The resource manager for an Initiative's indicators. 
        See :class:`~arcgis.apps.hub.IndicatorManager`.
        """
        return IndicatorManager(self._gis, self.item)
    
    def delete(self):
        """
        Deletes the initiative. If unable to delete, raises a RuntimeException.

        :return:
            A bool containing True (for success) or False (for failure). 

        .. code-block:: python

            USAGE EXAMPLE: Delete an initiative successfully

            initiative1 = myHub.initiatives.get('itemId12345')
            initiative1.delete()

            >> True
        """
        if self.item is not None:
            #Fetch Initiative Collaboration group
            _collab_groupId = self.item.properties['groupId']
            _collab_group = self._gis.groups.get(_collab_groupId)
            #Fetch Open Data Group
            _od_groupId = self.item.properties['openDataGroupId']
            _od_group = self._gis.groups.get(_od_groupId)
            #Disable delete protection on groups
            _collab_group.protected = False
            _od_group.protected = False
            #Delete groups and initiative
            _collab_group.delete()
            _od_group.delete()
            return self.item.delete()
    
    def update(self, initiative_properties=None, data=None, thumbnail=None, metadata=None):
        """ Updates the initiative.


        .. note::
            For initiative_properties, pass in arguments for only the properties you want to be updated.
            All other properties will be untouched.  For example, if you want to update only the
            initiative's description, then only provide the description argument in initiative_properties.


        =====================     ====================================================================
        **Argument**              **Description**
        ---------------------     --------------------------------------------------------------------
        initiative_properties     Required dictionary. See URL below for the keys and values.
        ---------------------     --------------------------------------------------------------------
        data                      Optional string. Either a path or URL to the data.
        ---------------------     --------------------------------------------------------------------
        thumbnail                 Optional string. Either a path or URL to a thumbnail image.
        ---------------------     --------------------------------------------------------------------
        metadata                  Optional string. Either a path or URL to the metadata.
        =====================     ====================================================================


        To find the list of applicable options for argument initiative_properties - 
        https://esri.github.io/arcgis-python-api/apidoc/html/arcgis.gis.toc.html#arcgis.gis.Item.update

        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python

            USAGE EXAMPLE: Update an initiative successfully

            initiative1 = myHub.initiatives.get('itemId12345')
            initiative1.update(initiative_properties={'description':'Create your own initiative to organize people around a shared goal.'})

            >> True
        """
        if initiative_properties:
            _initiative_data = self.definition
            for key, value in initiative_properties.items():
                _initiative_data[key] = value
            return self.item.update(_initiative_data, data, thumbnail, metadata)
    
class InitiativeManager(object):
    """
    Helper class for managing initiatives within a Hub. This class is not created by users directly. 
    An instance of this class, called 'initiatives', is available as a property of the Hub object. Users
    call methods on this 'initiatives' object to manipulate (add, get, search, etc) initiatives.
    """
    
    def __init__(self, hub, initiative=None):
        self._hub = hub
        self._gis = self._hub.gis
          
    def add(self, title, description=None, data=None, thumbnail=None):
        """ 
        Adds a new initiative to the Hub.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Required string.
        ---------------     --------------------------------------------------------------------
        description         Optional string. 
        ---------------     --------------------------------------------------------------------
        data                Optional string. Either a path or URL to the data.
        ---------------     --------------------------------------------------------------------
        thumbnail           Optional string. Either a path or URL to a thumbnail image.
        ===============     ====================================================================

        :return:
           The initiative if successfully added, None if unsuccessful.

        .. code-block:: python

            USAGE EXAMPLE: Add an initiative successfully

            initiative1 = myHub.initiatives.add(title='Vision Zero Analysis')
            initiative1.item
        """

        #Define initiative
        if description is None:
            description = 'Create your own initiative to organize people around a shared goal.'
        _item_dict = {"type":"Hub Initiative", "snippet":title + " Custom initiative", "typekeywords":"OpenData, Hub, hubInitiative", "title":title, "description": description, "licenseInfo": "CC-BY-SA","culture": "{{culture}}", "properties":{'schemaVersion':2}}
        
        #Defining open data and collaboration groups
        _od_group_title = title + ' Initiative Content Group'
        _od_group_dict = {"title": _od_group_title, "tags": ["Hub Initiative Group", "Open Data"], "access":"public", "isOpenData": True}
        _collab_group_title = title + ' Initiative Collaboration Group'
        _collab_group_dict = {"title": _collab_group_title, "tags": ["Hub Initiative Group", "initiativeCollaborationGroup"], "access":"org"}
        
        #Create groups
        od_group =  self._gis.groups.create_from_dict(_od_group_dict)
        collab_group =  self._gis.groups.create_from_dict(_collab_group_dict)
        
        #Protect groups from accidental deletion
        od_group.protected = True
        collab_group.protected = True
        
        #Adding it to _item_dict
        if od_group is not None and collab_group is not None:
            _item_dict['properties']['groupId'] = collab_group.id
            _item_dict['properties']['openDataGroupId'] = od_group.id
        
        #Create initiative and share it with collaboration group
        item =  self._gis.content.add(_item_dict, owner=self._gis.users.me.username)
        item.share(groups=[collab_group])
        
        #update initiative data
        _item_data = {"assets": [{"id": "bannerImage","url": self._hub.enterprise_org_url+"/sharing/rest/content/items/"+item.id+"/resources/detail-image.jpg","properties": {"type": "resource","fileName": "detail-image.jpg","mimeType": "image/jepg"},"license": {"type": "none"},"display": {"position": {"x": "center","y": "center"}}},{"id": "iconDark","url": self._hub.enterprise_org_url+"/sharing/rest/content/items/"+item.id+"/resources/icon-dark.png","properties": {"type": "resource","fileName": "icon-dark.png","mimeType": "image/png"},"license": {"type": "none"}},{"id": "iconLight","url": self._hub.enterprise_org_url+"/sharing/rest/content/items/"+item.id+"/resources/icon-light.png","properties": {"type": "resource","fileName": "icon-light.png","mimeType": "image/png"},"license": {"type": "none"}}],"steps": [{"id": "informTools","title": "Inform the Public","description": "Share data about your initiative with the public so people can easily find, download and use your data in different formats.","templateIds": [],"itemIds": []},{"id": "listenTools","title": "Listen to the Public","description": "Create ways to gather citizen feedback to help inform your city officials.","templateIds": [],"itemIds": []},{"id": "monitorTools","title": "Monitor Progress","description": "Establish performance measures that incorporate the publics perspective.","templateIds": [],"itemIds": []}],"indicators": [],"values": {"collaborationGroupId": collab_group.id,"openDataGroupId": od_group.id,"followerGroups": [],"bannerImage": {"source": "bannerImage","display": {"position": {"x": "center","y": "center"}}}}}
        _data = json.dumps(_item_data)
        item.update(item_properties={'text': _data})
        return Initiative(self._gis, item)
    
    def get(self, initiative_id):
        """ Returns the initiative object for the specified initiative_id.

        =======================    =============================================================
        **Argument**               **Description**
        -----------------------    -------------------------------------------------------------
        initiative_id              Required string. The initiative itemid.
        =======================    =============================================================

        :return:
            The initiative object if the item is found, None if the item is not found.

        .. code-block:: python

            USAGE EXAMPLE: Fetch an initiative successfully

            initiative1 = myHub.initiatives.get('itemId12345')
            initiative1.item

        """
        initiativeItem =    self._gis.content.get(initiative_id)
        if 'hubInitiative' in initiativeItem.typeKeywords:
            return Initiative(self._gis, initiativeItem)
        else:
            raise TypeError("Item is not a valid initiative or is inaccessible.")
    
    def search(self, scope=None, title=None, owner=None, created=None, modified=None, tags=None):
        """ 
        Searches for initiatives.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        scope               Optional string. Defines the scope of search.
                            Valid values are 'official', 'community' or 'all'.
        ---------------     --------------------------------------------------------------------
        title               Optional string. Return initiatives with provided string in title.
        ---------------     --------------------------------------------------------------------
        owner               Optional string. Return initiatives owned by a username.
        ---------------     --------------------------------------------------------------------
        created             Optional string. Date the initiative was created.
                            Shown in milliseconds since UNIX epoch.
        ---------------     --------------------------------------------------------------------
        modified            Optional string. Date the initiative was last modified.
                            Shown in milliseconds since UNIX epoch
        ---------------     --------------------------------------------------------------------
        tags                Optional string. User-defined tags that describe the initiative.
        ===============     ====================================================================

        :return:
           A list of matching initiatives.
        """

        initiativelist = []
        
        #Build search query
        query = 'typekeywords:hubInitiative'
        if title!=None:
            query += ' AND title:'+title
        if owner!=None:
            query += ' AND owner:'+owner
        if created!=None:
            query += ' AND created:'+created
        if modified!=None:
            query += ' AND modified:'+modified
        if tags!=None:
            query += ' AND tags:'+tags
        
        #Apply org scope and search
        if scope is None or self._gis.url=='https://www.arcgis.com':
            items = self._gis.content.search(query=query, max_items=5000)
        elif scope.lower()=='official':
            query += ' AND access:public'
            _gis = GIS(self._hub.enterprise_org_url)
            items = _gis.content.search(query=query, max_items=5000)
        elif scope.lower()=='community':
            query += ' AND access:public'
            _gis = GIS(self._hub.community_org_url)
            items = _gis.content.search(query=query, max_items=5000)
        elif scope.lower()=='all':
            items = self._gis.content.search(query=query, outside_org=True, max_items=5000)
        else:
            raise Exception("Invalid value for scope")
            
        #Return searched initiatives
        for item in items:
            initiativelist.append(Initiative(self._gis, item))
        return initiativelist
        
class Indicator(collections.OrderedDict):
    """
    Represents an indicator within an initiative. Initiatives use Indicators to standardize 
    data sources for ready-to-use analysis and comparison. Indicators are measurements of a system 
    including features, calculated metrics, or quantified goals. 
    """
    
    def __init__(self, initiativeItem, indicatorObject):
        """
        Constructs an empty Indicator object
        """
        self._initiativeItem = initiativeItem
        self._initiativedata = self._initiativeItem.get_data()
        self._indicatordict = indicatorObject
        pmap = PropertyMap(self._indicatordict)
        self.definition = pmap
            
    def __repr__(self):
        return '<%s id:"%s" optional:%s>' % (type(self).__name__, self.indicatorid, self.optional)
       
    @property
    def indicatorid(self):
        """
        Returns the id of the indicator
        """
        return self._indicatordict['id']
    
    @property
    def indicator_type(self):
        """
        Returns the type (Data/Parameter) of the indicator
        """
        return self._indicatordict['type']
    
    @property
    def optional(self):
        """
        Status if the indicator is optional (True/False)
        """
        return self._indicatordict['optional']
    
    @property
    def url(self):
        """
        Returns the data layer url (if configured) of the indicator
        """
        try:
            return self._indicatordict['source']['url']
        except:
            return 'Url not available for this indicator'
        
    @property
    def name(self):
        """
        Returns the layer name (if configured) of the indicator
        """
        try:
            return self._indicatordict['source']['url']
        except:
            return 'Name not available for this indicator'
        
    @property
    def itemid(self):
        """
        Returns the item id of the data layer (if configured) of the indicator
        """
        try:
            return self._indicatordict['source']['itemId']
        except:
            return 'Item Id not available for this indicator'
        
    @property
    def mappings(self):
        """
        Returns the attribute mapping from data layer (if configured) of the indicator
        """
        try:
            return self._indicatordict['source']['mappings']
        except:
            return 'Attribute mapping not available for this indicator'
    
    def delete(self):
        """
        Deletes an indicator from the initiative

        :return:
            A bool containing True (for success) or False (for failure). 

        .. code-block:: python

            USAGE EXAMPLE: Delete an indicator successfully

            indicator1 = initiative1.indicators.get('streetCrashes')
            indicator1.delete()

            >> True
        """
        if self._indicatordict is not None:
            _indicator_id = self._indicatordict['id']
            self._initiativedata['indicators'] = list(filter(lambda indicator: indicator.get('id')!=_indicator_id, self._initiativedata['indicators']))
            _new_initiativedata = json.dumps(self._initiativedata)
            return self._initiativeItem.update(item_properties={'text': _new_initiativedata})
     
    def get_data(self):
        """
        Retrieves the data associated with an indicator
        """
        return self.definition
    
    def update(self, indicator_properties=None):
        """
        Updates properties of an initiative

        :return:
            A bool containing True (for success) or False (for failure). 

        .. code-block:: python

            USAGE EXAMPLE: Update an indicator successfully

            indicator1_data = indicator1.get_data()
            indicator1_data['optional'] = False
            indicator1.update(indicator_properties = indicator1_data)

            >> True

            Refer the indicator definition (`get_data()`) to learn about fields that can be 
            updated and their acceptable data format.

        """
        try:
            _indicatorId = indicator_properties['id']
        except:
            return 'Indicator properties must include id of indicator'
        if indicator_properties is not None:
            self._initiativedata['indicators'] = [dict(indicator_properties) if indicator['id']==_indicatorId else indicator for indicator in self._initiativedata['indicators']]
            _new_initiativedata = json.dumps(self._initiativedata)
            status = self._initiativeItem.update(item_properties={'text': _new_initiativedata})      
            if status:
                self.definition = PropertyMap(indicator_properties)
                return status
    
class IndicatorManager(object):
    """Helper class for managing indicators within an initiative. This class is not created by users directly. 
    An instance of this class, called 'indicators', is available as a property of the Initiative object. Users
    call methods on this 'indicators' object to manipulate (add, get, search, etc) indicators of a particular
    initiative.
    """
    def __init__(self, gis, initiativeItem):
        self._gis = gis
        self._initiativeItem = initiativeItem
        self._initiativedata = self._initiativeItem.get_data()
        self._indicators = self._initiativedata['indicators']
        
    def add(self, indicator_properties):
        """
        Adds a new indicator to given initiative.

        *Key:Value Dictionary Options for Argument indicator_properties*

        =================  =====================================================================
        **Key**            **Value**
        -----------------  ---------------------------------------------------------------------
        id                 Required string. Indicator identifier within initiative template
        -----------------  ---------------------------------------------------------------------
        name               Optional string. Indicator name
        -----------------  ---------------------------------------------------------------------
        type               Optional string. Valid values are Data, Parameter.
        -----------------  ---------------------------------------------------------------------
        optional           Required boolean
        -----------------  ---------------------------------------------------------------------
        definition         Optional dictionary. Specification of the Indicator - types, fields
        -----------------  ---------------------------------------------------------------------
        source             Optional dictionary. Reference to an API or collection of data along 
                           with mapping between schemas
        =================  =====================================================================

        :return:
           A bool containing True (for success) or False (for failure).

        .. code-block:: python

            USAGE EXAMPLE: Add an indicator successfully

            indicator1_data = {'id': 'streetCrashes', 'type': 'Data', 'optional':False}
            initiative1.indicators.add(indicator_properties = indicator1_data)

            >> True

        """
        _stemplates = []
        _id = indicator_properties['id']
        _added = False
        
        #Fetch initiative template data
        _itemplateid = self._initiativedata['source']
        _itemplate =    self._gis.content.get(_itemplateid)
        _itemplatedata = _itemplate.get_data()
        
        #Fetch solution templates associated with initiative template
        for step in _itemplatedata['steps']:
            for _stemplateid in step['templateIds']:
                _stemplates.append(_stemplateid)
        
        #Fetch data for each solution template
        for _stemplateid in _stemplates:
            _stemplate =    self._gis.content.get(_stemplateid)
            _stemplatedata = _stemplate.get_data()
            
            #Check if indicator exists in solution
            for indicator in _stemplatedata['indicators']:
                
                #add indicator to initiative
                if indicator['id']==_id:
                    if self.get(_id) is not None:
                        return 'Indicator already exists'
                    else:
                        self._initiativedata['indicators'].append(indicator_properties)
                        _new_initiativedata = json.dumps(self._initiativedata)
                        self._initiativeItem.update(item_properties={'text': _new_initiativedata})
                        _added = True
                        return Indicator(self._initiativeItem, indicator_properties)
        if not _added:
            return 'Invalid indicator id for this initiative'
    
    def get(self, indicator_id):
        """ Returns the indicator object for the specified indicator_id.

        =======================    =============================================================
        **Argument**               **Description**
        -----------------------    -------------------------------------------------------------
        indicator_id               Required string. The indicator identifier.
        =======================    =============================================================

        :return:
            The indicator object if the indicator is found, None if the indicator is not found.

        """
        for indicator in self._indicators:
            if indicator['id']==indicator_id:
                _indicator = indicator
        try:
            return Indicator(self._initiativeItem, _indicator)
        except:
            return None
    
    def search(self, url=None, item_id=None, name=None):
        """ 
        Searches for indicators within an initiative.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        url                 Optional string. url registered for indicator in `source` dictionary.
        ---------------     --------------------------------------------------------------------
        item_id             Optional string. itemid registered for indicator in `source` dictionary.
        ---------------     --------------------------------------------------------------------
        name                Optional string. name registered for indicator in `source` dictionary.
        ===============     ====================================================================

        :return:
           A list of matching indicators.
        """
        _indicators = []
        indicatorlist = []
        for indicator in self._indicators:
            _indicators.append(indicator)
        if url!=None:
            _indicators = [indicator for indicator in _indicators if indicator['source']['url']==url]
        if item_id!=None:
            _indicators = [indicator for indicator in _indicators if indicator['source']['itemId']==item_id]
        if name!=None:
            _indicators = [indicator for indicator in _indicators if indicator['source']['name']==name]
        for indicator in _indicators:
            indicatorlist.append(Indicator(self._initiativeItem, indicator))
        return indicatorlist

class Event(collections.OrderedDict):
    """
    Represents an event in a Hub. A Hub has many Events that can be associated with an Initiative.
    Events are meetings for people to support an Initiative. Events are scheduled by an organizer 
    and have many attendees. An Event has a Group so that they can include content for preparation 
    as well as gather and archive content during the event for later retrieval or analysis.
    """
    def __init__(self, gis, eventObject):
        """
        Constructs an empty Event object
        """
        self._gis = gis
        self._eventgeometry = eventObject.geometry
        self._eventdict = eventObject.attributes
        pmap = PropertyMap(self._eventdict)
        self.definition = pmap
            
    def __repr__(self):
        return '<%s title:"%s" location:%s>' % (type(self).__name__, self.title, self.location)
    
    @property
    def title(self):
        """
        Returns the title of the event
        """
        return self._eventdict['title']
    
    @property
    def location(self):
        """
        Returns the location of the event
        """
        return self._eventdict['location'] 
    
    @property
    def initiativeid(self):
        """
        Returns the initiative id of the event if it belongs to an Initiative
        """
        return self._eventdict['initiativeId'] 
    
    @property
    def siteid(self):
        """
        Returns the site id of the event site
        """
        return self._eventdict['siteId']
    
    @property
    def organizer_name(self):
        """
        Returns the name of the organizer of the event
        """
        return self._eventdict['organizerName'] 
    
    @property
    def organizers(self):
        """
        Returns names of all organizers of the event
        """
        return self._eventdict['organizers']
    
    @property
    def description(self):
        """
        Returns description of the event
        """
        return self._eventdict['description']
    
    @property
    def start_date(self):
        """
        Returns start date of the event in milliseconds since UNIX epoch
        """
        return self._eventdict['startDate']
    
    @property
    def end_date(self):
        """
        Returns end date of the event in milliseconds since UNIX epoch
        """
        return self._eventdict['endDate']
    
    @property
    def creator(self):
        """
        Returns creator of the event
        """
        return self._eventdict['Creator']
    
    @property
    def capacity(self):
        """
        Returns attendance capacity for attendees of the event
        """
        return self._eventdict['capacity']
    
    @property
    def attendance(self):
        """
        Returns attendance count for a past event
        """
        return self._eventdict['attendance']
    
    @property
    def status(self):
        """
        Returns status of the event
        """
        return self._eventdict['status']
    
    @property
    def is_cancelled(self):
        """
        Check if event is Cancelled
        """
        return self._eventdict['isCancelled']
    
    @property
    def geometry(self):
        """
        Returns co-ordinates of the event location
        """
        return self._eventgeometry
    
class EventManager(object):
    """Helper class for managing events within a Hub. This class is not created by users directly. 
    An instance of this class, called 'events', is available as a property of the Hub object. Users
    call methods on this 'events' object to manipulate (add, search, get_map etc) events 
    of a particular Hub. 
    """
    def __init__(self, hub, event=None):
        self._gis = hub.gis
        if event:
            self._event = event
            
    def __all_events(self):
        """
        Fetches all events for particular hub.
        """
        events = []
        _events_layer = self._gis.content.search(query="typekeywords:hubEventsLayer", max_items=5000)[0]
        _events_layer_url = _events_layer.url + '/0'
        _events_data = FeatureLayer(_events_layer_url).query().features
        for event in _events_data:
            events.append(Event(self._gis, event))
        return events
    
    def search(self, initiative_id=None, title=None, location=None, organizer_name=None):
        """ 
        Searches for events within a Hub.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        initiative_id       Optional string. Initiative itemid.
        ---------------     --------------------------------------------------------------------
        title               Optional string. Title of the event.
        ---------------     --------------------------------------------------------------------
        location            Optional string. Location where event is held.
        ---------------     --------------------------------------------------------------------
        organizer_name      Optional string. Name of the organizer of the event.
        ===============     ====================================================================

        :return:
           A list of matching indicators.
        """
        events = []
        events = self.__all_events()
        if initiative_id!=None:
            events = [event for event in events if initiative_id==event.initiativeid]
        if title!=None:
            events = [event for event in events if title in event.title]
        if location!=None:
            events = [event for event in events if location in event.location]
        if organizer_name!=None:
            events = [event for event in events if organizer_name==event.organizer_name]
        return events
    
    def get_map(self):
        """
        Plot all events for a Hub in an embedded webmap within the notebook.
        """
        _events_layer = self._gis.content.search(query="typekeywords:hubEventsLayer", max_items=5000)[0]
        event_map = self._gis.map(zoomlevel=2)
        event_map.basemap = 'dark-gray'
        event_map.add_layer(_events_layer, {'title':'Event locations for this Hub','opacity':0.7})
        return event_map
